###############################################################################
#
#   Copyright: (c) 2015 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

from onyx.core import FloatField, ValueType, Archivable
from onyx.core import MktIndirectionFactory, EnforceArchivableEntitlements

__all__ = ["PriceFix"]


###############################################################################
@EnforceArchivableEntitlements("Database", "ArchivedOverwritable")
class PriceFix(Archivable):
    """
    Class used to represent price fixes (marks or settlement values).
    """
    # -------------------------------------------------------------------------
    @MktIndirectionFactory(FloatField)
    def Price(self, graph):
        pass

    # -------------------------------------------------------------------------
    @ValueType()
    def LastKnot(self, graph):
#        date = date or graph("Database", "MktDataDate")
        date = graph("Database", "MktDataDate")
        return self.get_dated("Price", date, strict=False)

    # -------------------------------------------------------------------------
    @ValueType("PropSubGraph")
    def PrcFixCurve(self, graph, start=None, end=None):
        return self.get_history("Price", start, end)


# -----------------------------------------------------------------------------
def prepare_for_test():
    from . import ufo_database
    ufo_database.prepare_for_test()



###############################################################################
#
#   Copyright: (c) 2015 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

from onyx.core import Curve, RDate, CurveIntersect
from onyx.core import UfoBase, GetVal, ValueType
from onyx.core import ReferenceField, StringField, FloatField

from ..econometrics.regression import RobustRegression
from ..econometrics.arma_functions import AR1

import numpy as np

__all__ = ["EquityPair"]


# -----------------------------------------------------------------------------
def get_spread_parms(crv1, crv2):
    x = np.log(crv2.values)
    y = np.log(crv1.values)

    m, c, snr, spr = RobustRegression(x, y)

    root, std_err, cv, ar1_res = AR1(spr, const=True)

    return {"m": m, "c": c, "snr": snr,
            "root": root, "std_err": std_err, "cv": cv}


###############################################################################
class FitError(Exception):
    pass


###############################################################################
class EquityPair(UfoBase):
    """
    class used to represent a pair of two equities.
    """
    Symbol1 = ReferenceField(obj_type="EquityAsset")
    Symbol2 = ReferenceField(obj_type="EquityAsset")

    # --- parameters used to define a spread
    RegRange = StringField("-24m")
    Alpha = FloatField()
    Beta = FloatField()

    name_fmt = "EQP {0:s} {1:s} - {2:s} {3:s}"

    # -------------------------------------------------------------------------
    def __post_init__(self):
        sym1 = GetVal(self.Symbol1, "Symbol")
        sym2 = GetVal(self.Symbol2, "Symbol")
        exchg_code1 = GetVal(GetVal(self.Symbol1, "Exchange"), "Code")
        exchg_code2 = GetVal(GetVal(self.Symbol2, "Exchange"), "Code")

        self.Name = self.name_fmt.format(sym1, exchg_code1, sym2, exchg_code2)

    # -------------------------------------------------------------------------
    @ValueType("PropSubGraph")
    def Curves(self, graph, start=None, end=None):
        sym1 = graph(self, "Symbol1")
        sym2 = graph(self, "Symbol2")
        ccy1 = graph(sym1, "Denominated")
        ccy2 = graph(sym2, "Denominated")

        sym1_crv = graph(sym1, "GetCurve", start, end, "Close")
        sym2_crv = graph(sym2, "GetCurve", start, end, "Close")

        if ccy1 != ccy2:
            # --- convert both to US$
            cross1 = "{0:3s}/USD".format(ccy1)
            cross2 = "{0:3s}/USD".format(ccy2)
            cross1_crv = graph(cross1, "GetCurve", start, end)
            cross2_crv = graph(cross2, "GetCurve", start, end)

            sym1_crv, sym2_crv, cross1_crv, cross2_crv = CurveIntersect([
                sym1_crv, sym2_crv, cross1_crv, cross2_crv])

            sym1_crv *= cross1_crv
            sym2_crv *= cross2_crv

            return sym1_crv, sym2_crv

        else:
            return CurveIntersect([sym1_crv, sym2_crv])

    # -------------------------------------------------------------------------
    @ValueType("Callable")
    def Ratio(self, graph, start, end, use_log=False):
        """
        Description:
            Ratio is defined as:
                Ratio = Symbol1 / Symbol2
            or, if use_log is True:
                Ratio = log(Symbol1 / Symbol2)
        """
        num_crv, den_crv = graph(self, "Curves", start, end)

        if use_log:
            return Curve(num_crv.dates,
                         np.log(num_crv.values / den_crv.values))
        else:
            return Curve(num_crv.dates, num_crv.values / den_crv.values)

    # -------------------------------------------------------------------------
    @ValueType("Callable")
    def Spread(self, graph, start, end):
        """
        Description:
            Spread is defined as:
                Spread = log(Symbol1) - Beta*log(Symbol2) - Alpha
            NB: Alpha and Beta are the stored parameters.
        Inputs:
            start - spread start date
            end   - spread end date
        """
        sym1_crv, sym2_crv = graph(self, "Curves", start, end)

        c = graph(self, "Alpha")
        m = graph(self, "Beta")

        spread = np.log(sym1_crv.values) - m*np.log(sym2_crv.values) - c

        return Curve(sym1_crv.dates, spread)

    # -------------------------------------------------------------------------
    @ValueType("Callable")
    def Ar1Parms(self, graph, start, end):
        """
        Returns:
            A tuple: (root, std_err, cv, ar1_res)
        """
        return AR1(graph(self, "Spread", start, end), const=True)

    # -------------------------------------------------------------------------
    @ValueType("Callable")
    def Ar1Root(self, graph, start, end):
        parms = graph(self, "Ar1Parms", start, end)
        return parms[0]

    # -------------------------------------------------------------------------
    def fit_spread_parms(self, min_range=None, max_range=None, cv_thr=1.0):
        """
        Description:
            Determine the longest regression range for which we can reject the
            unit root hypothesis for the spread determined via robust
            regression.
        Inputs:
            min_range - minimum value of regression range, in number of months
            max_range - maximum value of regression range, in number of months
            cv_thr    - critical value threshold:
                            the range is accepted if cv > cv_thr
        """
        min_range = min_range or 6
        max_range = max_range or 36

        sym1_crv, sym2_crv = GetVal(self, "Curves")
        range_ed = GetVal("Database", "PricingDate")

        for mths in range(max_range, min_range - 1, -3):
            rule = "-{0:d}m".format(mths)
            range_sd = range_ed + RDate(rule)

            crv1 = sym1_crv.crop(range_sd, range_ed)
            crv2 = sym2_crv.crop(range_sd, range_ed)
            parms = get_spread_parms(crv1, crv2)

            # --- spread is accepted if cv > 1.0
            if parms["cv"] > cv_thr:
                parms["rule"] = rule
                return parms

        raise FitError("{0:s} is never cointegrated for any range in "
                       "the last {1:d} months".format(self.Name, max_range))

    # -------------------------------------------------------------------------
    def analyze_spread_parms(self, min_range=None, max_range=None):

        min_range = min_range or 6
        max_range = max_range or 48

        sym1_crv, sym2_crv = GetVal(self, "Curves")
        range_ed = GetVal("Database", "PricingDate")

        head_fmt = "{0:>6s}  {1:>6s}  {2:>6s}  {3:>6s}  {4:>6s}"
        body_fmt = "{rule:>6s}  {m:6.3f}  {c:6.3f}  {root:6.4f}  {cv:6.3f}"
        dashes = "-"*6

        print(head_fmt.format("rule", "m", "c", "root", "cv"))
        print(head_fmt.format(dashes, dashes, dashes, dashes, dashes))

        for mths in range(max_range, min_range - 1, -3):
            rule = "-{0:d}m".format(mths)
            range_sd = range_ed + RDate(rule)

            crv1 = sym1_crv.crop(range_sd, range_ed)
            crv2 = sym2_crv.crop(range_sd, range_ed)
            parms = get_spread_parms(crv1, crv2)
            parms["rule"] = rule

            print(body_fmt.format(**parms))

    # -------------------------------------------------------------------------
    def set_spread_parms(self, reg_range=None):
        if reg_range is None:
            parms = self.fit_spread_parms()
        else:
            rule = "-{0:d}m".format(reg_range)
            sym1_crv, sym2_crv = GetVal(self, "Curves")
            range_ed = GetVal("Database", "PricingDate")
            range_sd = range_ed + RDate(rule)
            crv1 = sym1_crv.crop(range_sd, range_ed)
            crv2 = sym2_crv.crop(range_sd, range_ed)
            parms = get_spread_parms(crv1, crv2)

        self.Alpha = parms["c"]
        self.Beta = parms["m"]
        self.RegRange = rule


# -----------------------------------------------------------------------------
def prepare_for_test():
    from onyx.core import AddIfMissing

    from .ufo_equity_asset import EquityAsset

    from ..system import ufo_currency
    from ..system import ufo_database
    from ..system import ufo_holiday_calendar
    from ..system import ufo_exchange

    ufo_database.prepare_for_test()
    ufo_holiday_calendar.prepare_for_test()
    ufo_currency.prepare_for_test()
    ufo_exchange.prepare_for_test()

    sse_info = {
        "Symbol": "SSE",
        "Exchange": "XLON Exchange",
        "Tickers": {
            "Bloomberg": "SSE",
        },
        "Multiplier": 0.01,
    }

    cna_info = {
        "Symbol": "CNA",
        "Exchange": "XLON Exchange",
        "Tickers": {
            "Bloomberg": "CNA",
        },
        "Multiplier": 0.01,
    }

    AddIfMissing(EquityAsset(**sse_info))
    AddIfMissing(EquityAsset(**cna_info))
    AddIfMissing(EquityPair(Symbol1="EQ SSE LN", Symbol2="EQ CNA LN"))



###############################################################################
#
#   Copyright: (c) 2015 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

from onyx.core import Date, UfoBase, ValueType

__all__ = ["Database"]


###############################################################################
class Database(UfoBase):
    # -------------------------------------------------------------------------
    @ValueType()
    def PricingDate(self, graph):
        return Date.today()

    # -------------------------------------------------------------------------
    @ValueType()
    def MktDataDate(self, graph):
        # --- defaults to PricingDate
        return graph(self, "PricingDate")

    # -------------------------------------------------------------------------
    @ValueType()
    def PositionsDate(self, graph):
        """
        Used to retrive children by book.
        """
        # --- defaults to PricingDate
        return graph(self, "PricingDate")

    # -------------------------------------------------------------------------
    @ValueType()
    def ProcDate(self, graph):
        return Date.today()

    # -------------------------------------------------------------------------
    @ValueType()
    def ReportDate(self, graph):
        return Date.today()

    # -------------------------------------------------------------------------
    @ValueType()
    def ForceStrict(self, graph):
        """
        Change to enforce retrival of a specific dated object or knot.
        """
        return False

    # -------------------------------------------------------------------------
    @ValueType()
    def ArchivedOverwritable(self, graph):
        """
        Change to allow overwriting archived objects.
        """
        return False

    # -------------------------------------------------------------------------
    @ValueType()
    def TradesDeletable(self, graph):
        """
        Change to allow deleting trades.
        """
        return False


# -----------------------------------------------------------------------------
def prepare_for_test():
    from onyx.core import AddIfMissing
    AddIfMissing(Database(Name="Database"))



###############################################################################
#
#   Copyright: (c) 2015 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

from onyx.core import ObjNamesByType, ValueType
from onyx.core import UfoBase, StringField, ListField

__all__ = ["Category"]


###############################################################################
class Category(UfoBase):
    """
    Class used to collect instances of classes listed in ObjTypes that match
    the following pattern:
        graph(instance, "ValueType") == graph(self, "TargetValue")
    """
    ObjTypes = ListField()
    TargetVT = StringField()
    TargetValue = StringField()

    # -------------------------------------------------------------------------
    def __post_init__(self):
        self.TargetValue = self.TargetValue or self.Name

    # -------------------------------------------------------------------------
    @ValueType()
    def Items(self, graph):
        vt = graph(self, "TargetVT")
        value = graph(self, "TargetValue")
        types = graph(self, "ObjTypes")
        names = [name for obj_type in types
                 for name in ObjNamesByType(obj_type)]
        return sorted([name for name in names if graph(name, vt) == value])


# -----------------------------------------------------------------------------
def prepare_for_test():
    from onyx.core import AddIfMissing

    AddIfMissing(Category(Name="Nth. America",
                          ObjTypes=["EquityAsset"], TargetVT="Region"))
    AddIfMissing(Category(Name="Nth. Europe",
                          ObjTypes=["EquityAsset"], TargetVT="Region"))
    AddIfMissing(Category(Name="Cnt. Europe",
                          ObjTypes=["EquityAsset"], TargetVT="Region"))

    AddIfMissing(Category(Name="USA",
                          ObjTypes=["EquityAsset"], TargetVT="Country"))
    AddIfMissing(Category(Name="United Kingdom",
                          ObjTypes=["EquityAsset"], TargetVT="Country"))
    AddIfMissing(Category(Name="Germany",
                          ObjTypes=["EquityAsset"], TargetVT="Country"))



###############################################################################
#
#   Copyright: (c) 2015 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

from onyx.core import unique_id, Structure
from onyx.core import StructureField, ReferenceField, UfoBase, ValueType

from ..risk.decorators import WithRiskValueTypes

__all__ = ["ShadowBook"]


###############################################################################
@WithRiskValueTypes
class ShadowBook(UfoBase):
    """
    Class used to represent a shadow book (i.e. a book whose children are not
    generated from trades and positions, but set manually).
    """
    Denominated = ReferenceField(obj_type="Currency")
    Children = StructureField(default=Structure())

    # -------------------------------------------------------------------------
    def __post_init__(self):
        self.Name = self.Name or "SB {0:s}".format(ShadowBook.random_name(16))

    # -------------------------------------------------------------------------
    @ValueType()
    def Leaves(self, graph):
        leaves = Structure()
        for child, qty in graph(self, "Children").items():
            leaves += qty*graph(child, "Leaves")
        return leaves

    # -------------------------------------------------------------------------
    @ValueType()
    def MktValUSD(self, graph):
        mtm = 0.0
        for leaf, qty in graph(self, "Leaves").items():
            mtm += qty*graph(leaf, "MktValUSD")
        return mtm

    # -------------------------------------------------------------------------
    @ValueType()
    def MktVal(self, graph):
        denominated = graph(self, "Denominated")
        spot_fx = graph("{0:3s}/USD".format(denominated), "Spot")
        return graph(self, "MktValUSD") / spot_fx

    # -------------------------------------------------------------------------
    @classmethod
    def random_name(cls, nchar=8):
        return unique_id(nchar)



###############################################################################
#
#   Copyright: (c) 2015 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

import numpy as np

__all__ = ["GbmPricePaths"]


# -----------------------------------------------------------------------------
def GbmPricePaths(vol, n_days, rd=0.0, rf=0.0, n_sims=1000, acc=False):
    """
    Description:
        Returns normalized simulated price paths. The actual prices can be
        obtained by multiplying the paths by the current spot price.
    Inputs:
        vol    - the daily volatility
        n_days - number of days in the simulation. each path will be of length
                 ndays+1, with initial value set to 1.
        rd     - daily domestic risk-free interest rate
        rf     - daily foreign risk-free interest rate (or dividend yield)
        n_sims - number of simulation paths
        acc    - if True, it will only return the final normalized prices after
                 n_days days
    Returns:
        numpy.array, according to acc
    """
    # --- Ito's drift term: it's only present when using continuous compounding
    #     of daily returns, i.e. using e^(r*n) instead of (1+r)^n
    ito = 0.5*vol*vol

    if acc:
        half = int(n_sims / 2)
        n_sims = 2*half
        prices = np.ones(n_sims)
        for n in range(n_days):
            z = np.random.standard_normal(half)
            z = np.hstack((z, -z))
            prices *= np.exp(rd - rf - ito + vol*z)

        return prices

    else:
        half = int(n_sims / 2)
        z = np.random.standard_normal((n_days, half))
        drift = rd - rf - ito

        paths = np.ones((n_days + 1, n_sims))
        paths[1:,:] = np.hstack((np.cumprod(np.exp(drift - vol*z), axis=0),
                                 np.cumprod(np.exp(drift + vol*z), axis=0)))

        return paths



###############################################################################
#
#   Copyright: (c) 2015 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

from onyx.core import ValueType, ReferenceField, DateField

from .tradable_api import AgingTradableObj, AddByInference
from .ufo_cash_balance import CashBalance


###############################################################################
class ForwardCash(AgingTradableObj):
    """
    Tradable class that represents cash to be exchanged at a future date.
    """
    Currency = ReferenceField(obj_type="Currency")
    PaymentDate = DateField()

    # -------------------------------------------------------------------------
    @ValueType()
    def MktVal(self, graph):
        pd = graph("Database", "PricingDate")
        ed = graph(self, "PaymentDate")

        if pd < ed:
            ccy = graph(self, "Currency")
            return graph(ccy, "DiscountFactor", ed, pd)
        else:
            return 1.0

    # -------------------------------------------------------------------------
    @ValueType()
    def MktValUSD(self, graph):
        cross = "{0:3s}/USD".format(graph(self, "Currency"))
        return graph(self, "MktVal")*graph(cross, "Spot")

    # -------------------------------------------------------------------------
    @ValueType()
    def ExpirationDate(self, graph):
        return graph(self, "PaymentDate")

    # -------------------------------------------------------------------------
    @ValueType()
    def NextTransactionDate(self, graph):
        return graph(self, "PaymentDate")

    # -------------------------------------------------------------------------
    @ValueType()
    def TradeTypes(self, graph):
        mapping = super().TradeTypes
        mapping.update({
            "Pay/Receive": "PayReceiveSecurities",
        })
        return mapping

    # -------------------------------------------------------------------------
    @ValueType()
    def ExpectedTransaction(self, graph):
        return "Pay/Receive"

    # -------------------------------------------------------------------------
    @ValueType()
    def PayReceiveSecurities(self, graph):
        cash = CashBalance(Currency=graph(self, "Currency"))
        return [
            {"Security": AddByInference(cash, in_memory=True),
             "Quantity": 1.0},
        ]

    # -------------------------------------------------------------------------
    @property
    def ImpliedName(self):
        pay_date = self.PaymentDate.strftime("%d%m%y")
        return "FWD {0:3s} {1:6s}".format(self.Currency, pay_date)


# -----------------------------------------------------------------------------
def prepare_for_test():
    from onyx.core import Date, RDate
    from .tradable_api import AddByInference

    from . import ufo_database
    from . import ufo_holiday_calendar
    from . import ufo_currency
    from . import ufo_currency_cross

    ufo_database.prepare_for_test()
    ufo_holiday_calendar.prepare_for_test()
    ufo_currency.prepare_for_test()
    ufo_currency_cross.prepare_for_test()

    paydt = Date.today() + RDate("+1y")

    securities = []
    for ccy in ["USD", "EUR", "GBP"]:
        securities.append(AddByInference(ForwardCash(Currency=ccy,
                                                     PaymentDate=paydt)))

    return [sec.Name for sec in securities]



###############################################################################
#
#   Copyright: (c) 2015 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

from onyx.core import ReferenceField, StringField, FloatField
from onyx.core import ValueType, Archivable
from onyx.core import MktIndirectionFactory, EnforceArchivableEntitlements

from ..risk.decorators import WithRiskValueTypes

__all__ = ["Fund"]


###############################################################################
@WithRiskValueTypes
@EnforceArchivableEntitlements("Database", "ArchivedOverwritable")
class Fund(Archivable):
    """
    Class used to represent an investment Fund.
    """
    # --- base currency for the whole fund
    Denominated = ReferenceField(obj_type="Currency")
    # --- this portfolio captures all investments
    Portfolio = ReferenceField(obj_type="Portfolio")
    # --- this book captures subscriptions/redemptions as well as costs and
    #     amortized costs
    CashAccount = ReferenceField(obj_type="Book")
    DisplayName = StringField()

    # -------------------------------------------------------------------------
    @ValueType()
    def Leaves(self, graph):
        port_leaves = graph(graph(self, "Portfolio"), "Leaves")
        cash_leaves = graph(graph(self, "CashAccount"), "Leaves")
        return port_leaves + cash_leaves

    # -------------------------------------------------------------------------
    @ValueType()
    def MktValUSD(self, graph):
        port_mtm = graph(graph(self, "Portfolio"), "MktValUSD")
        cash_mtm = graph(graph(self, "CashAccount"), "MktValUSD")
        return port_mtm + cash_mtm

    # -------------------------------------------------------------------------
    @ValueType()
    def MktVal(self, graph):
        denominated = graph(self, "Denominated")
        spot_fx = graph("{0:3s}/USD".format(denominated), "Spot")
        return graph(self, "MktValUSD") / spot_fx

    # -------------------------------------------------------------------------
    @MktIndirectionFactory(FloatField)
    def Nav(self, graph):
        pass

    # -------------------------------------------------------------------------
    @MktIndirectionFactory(FloatField)
    def Gav(self, graph):
        pass

    # -------------------------------------------------------------------------
    @MktIndirectionFactory(FloatField)
    def Aum(self, graph):
        pass

    # -------------------------------------------------------------------------
    @MktIndirectionFactory(FloatField)
    def NumberOfShares(self, graph):
        pass

    # -------------------------------------------------------------------------
    @ValueType("PropSubGraph")
    def NavCurve(self, graph, start=None, end=None):
        return self.get_history("Nav", start, end)

    # -------------------------------------------------------------------------
    @ValueType("PropSubGraph")
    def GavCurve(self, graph, start=None, end=None):
        return self.get_history("Gav", start, end)

    # -------------------------------------------------------------------------
    @ValueType("PropSubGraph")
    def AumCurve(self, graph, start=None, end=None):
        return self.get_history("Aum", start, end)



###############################################################################
#
#   Copyright: (c) 2015 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

from onyx.core import Date, Date2LYY, LYY2Date
from onyx.core import CountBizDays, DateOffset, DateRange
from onyx.core import Curve, CurveIntersect
from onyx.core import TsNotFound
from onyx.core import GetVal

import numpy as np
import collections


# -----------------------------------------------------------------------------
def get_strip(commod, start, end, avg_type="LAST"):
    """
    Description:
        Return the historical timeseries for the required strip. The strip
        is defined as the average of all contracts with delivery date
        within start and end dates.
    Inputs:
        commod   - the name of a CommodAsset
        start    - the start date defining the strip
        end      - the   end date defining the strip
        avg_type - the averaging type. Valid choices are:
                    - DAILYCALDAYS
                    - DAILYBIZDAYS
                    - LAST
    Example:
        GetStrip(Date(2012, 1, 1), Date(2012, 12, 31), "DAILYBIZDAYS")
        returns the historical curve for the Cal12 swap for a commodity
        that prices out every business day.
    """
    cal = GetVal(commod, "HolidayCalendar")
    sdr = GetVal(commod, "SettDateRule")
    nbo = GetVal(commod, "NrbyOffset")

    avg_type = avg_type.upper()
    if avg_type == "DAILYCALDAYS":
        sd_rule = "+0d"
        ed_rule = "-0d"
        inc_rule = "+1d"
    elif avg_type == "DAILYBIZDAYS":
        sd_rule = "+0b"
        ed_rule = "-0b"
        inc_rule = "+1b"
    elif avg_type == "LAST":
        sd_rule = "+0J{0:s}".format(sdr)
        ed_rule = "+0J{0:s}".format(sdr)
        inc_rule = "+{0:d}m+0J{1:s}".format(1+nbo, sdr)
    else:
        raise NameError("Unrecognized "
                        "averaging type {0:s}".format(avg_type))

    sd = DateOffset(start, sd_rule, cal)
    ed = DateOffset(end, ed_rule, cal)
    cnts = [GetVal(commod, "ActiveByDate", d)
            for d in DateRange(sd, ed, inc_rule, cal)]

    # --- accumulate historical prices and quantities for all relevant
    #     contracts
    crvs, qtys = [], []
    for cnt, qty in collections.Counter(cnts).items():
        cnt = GetVal(commod, "GetContract", cnt)
        crvs.append(GetVal(cnt, "GetCurve", field="close"))
        qtys.append(float(qty))

    crvs = CurveIntersect(crvs)
    qtot = qtys[0]
    svls = qtys[0]*crvs[0].values  # sum of values (element by element)
    for k in range(1, len(crvs)):
        qtot += qtys[k]
        svls += qtys[k]*crvs[k].values

    return Curve(crvs[0].dates, svls/qtot)


# -----------------------------------------------------------------------------
def fwd_curve(commod):
    """
    Description:
        Return the forward curve as of a given date.
    Inputs:
        commod - the name of a CommodAsset
    Returns:
        A curve.
    """
    mdd = GetVal("Database", "MktDataDate")
    cal = GetVal(commod, "HolidayCalendar")
    if mdd <= DateOffset(mdd, GetVal(commod, "SettDateRule"), cal):
        offset = GetVal(commod, "NrbyOffset")
    else:
        offset = 1 + GetVal(commod, "NrbyOffset")

    sd = DateOffset(mdd, "+{0:d}m+0J".format(offset), cal)

    fwd_crv = Curve()
    for del_mth in GetVal(commod, "Contracts"):
        # --- contract's month start date
        cnt_sd = LYY2Date(del_mth)
        # --- only include contracts in the range
        if sd <= cnt_sd:
            cnt = GetVal(commod, "GetContract", del_mth)
            try:
                fwd_crv[cnt_sd] = GetVal(cnt, "Spot")
            except TsNotFound:
                # --- no time series for this contract: skip
                continue

    return fwd_crv


# -----------------------------------------------------------------------------
def nearby(commod, idx=0, start=None, end=None, smooth=False):
    """
    Description:
        Returns the historical time series for a given nearby month.
    Inputs:
        commod - the name of a CommodAsset
        idx    - the nearby index (0 is prompt month)
        start  - the start date of the curve
        end    - the   end date of the curve
        smooth - if True, use a weighted combination of two consecutive
                 months
    """
    if idx < 0:
        raise ValueError("Nearby index cannot "
                         "be negative: idx = {0:d}".format(idx))

    # --- if available, use default values for sd and ed
    start = start or Date.low_date()
    end = end or GetVal("Database", "MktDataDate")
    cal = GetVal(commod, "HolidayCalendar")

    # --- date rules that define the portion of the curve to be used for
    #     each contract
    rule_sd = "-1m{0:s}+1d".format(GetVal(commod, "SettDateRule"))
    rule_ed = GetVal(commod, "SettDateRule")

    # --- for all contracts in the range, select the ones that have a
    #     non-empty time series within the required date range
    cnts_iter = (Date2LYY(d) for d in DateRange(start, end, "0J+1m", cal))
    cnts_within_range = [cnt for cnt in cnts_iter
                         if cnt in GetVal(commod, "Contracts")]

    # --- get start dates for the first and last contracts in the range
    first_cnt_sd = LYY2Date(cnts_within_range[0])
    last_cnt_sd = LYY2Date(cnts_within_range[-1])
    last_cnt_sd = DateOffset(last_cnt_sd,
                             "{0:d}m+0J".format(GetVal(commod, "NrbyOffset")))

    dts, vls = [], []
    if smooth:
        # --- use a weighted combination of two consecutive months. weights
        #     are scaled linearly.
        rule0 = "{0:d}m+0J".format(idx)
        rule1 = "{0:d}m+0J".format(1+idx)
        for d in DateRange(first_cnt_sd, last_cnt_sd, "+1m", cal):
            mth0 = DateOffset(d, rule0, cal)
            mth1 = DateOffset(d, rule1, cal)
            csd = DateOffset(d, rule_sd, cal)
            ced = DateOffset(d, rule_ed, cal)
            try:
                cnt0 = GetVal(commod, "GetContract", Date2LYY(mth0))
                cnt1 = GetVal(commod, "GetContract", Date2LYY(mth1))
                crv0 = GetVal(cnt0, "GetCurve", csd, ced, field="close")
                crv1 = GetVal(cnt1, "GetCurve", csd, ced, field="close")
            except TsNotFound:
                continue

            m = CountBizDays(csd, ced, cal)
            n = len(crv0)

            if n != m or n != len(crv1):
                continue

            w = (np.cumsum(np.ones(n)) - 1) / (n - 1)
            mix = (1.0 - w)*crv0.values + w*crv1.values

            dts += crv0.dates
            vls += mix.tolist()
    else:
        rule = "{0:d}m+0J".format(idx)
        for d in DateRange(first_cnt_sd, last_cnt_sd, "+1m", cal):
            mth = DateOffset(d, rule, cal)
            csd = DateOffset(d, rule_sd, cal)
            ced = DateOffset(d, rule_ed, cal)
            try:
                cnt = GetVal(commod, "GetContract", Date2LYY(mth))
                crv = GetVal(cnt, "GetCurve", csd, ced, field="close")
                dts += crv.dates
                vls += crv.values.tolist()
            except TsNotFound:
                continue

    return Curve(dts, vls).crop(start, end)


# -----------------------------------------------------------------------------
def rolling_strip(commod, strip="Quarter",
                  idx=1, start=None, end=None, avg_type="LAST"):
    """
    Description:
        Return the historical time series for a given rolling strip.
    Inputs:
        commod   - the name of a CommodAsset
        strip    - choose among "Quarter", "Calendar"
        idx      - the nearby index (0 is prompt strip)
        start    - the start date of the curve
        end      - the end date of the curve
        avg_type - the averaging type (see get_strip).
    Returns:
        A Curve.
    """
    if idx < 0:
        raise ValueError("Nearby index cannot "
                         "be negative: idx = {0:d}".format(idx))

    # --- if available, use default values for sd and ed
    start = start or Date.low_date()
    end = end or GetVal("Database", "MktDataDate")
    cal = GetVal(commod, "HolidayCalendar")

    # --- date rules that define the portion of the curve to be used for
    #     each contract
    rule_sd = "-1m{0:s}+1d".format(GetVal(commod, "SettDateRule"))
    rule_ed = GetVal(commod, "SettDateRule")

    if strip == "Quarter":
        rule_i = "+{0:d}m+q".format(3*idx)
        rule_f = "+{0:d}m+Q".format(3*idx)
    elif strip == "Calendar":
        rule_i = "+{0:d}y+A".format(idx)
        rule_f = "+{0:d}y+E".format(idx)
    else:
        raise RuntimeError("Unrecognized strip type {0:s}".format(strip))

    # --- for all contracts in the range, select the ones that have a
    #     non-empty time series within the required date range
    cnts_iter = (Date2LYY(d) for d in DateRange(start, end, "0J+1m", cal))
    cnts_within_range = [cnt for cnt in cnts_iter
                         if cnt in GetVal(commod, "Contracts")]

    # --- get start dates for the first and last contracts in the range
    first_cnt_sd = LYY2Date(cnts_within_range[0])
    last_cnt_sd = LYY2Date(cnts_within_range[-1])
    last_cnt_sd = DateOffset(last_cnt_sd,
                             "{0:d}m+0J".format(GetVal(commod, "NrbyOffset")))

    dts, vls = [], []
    for d in DateRange(first_cnt_sd, last_cnt_sd, "+1m", cal):
        di = DateOffset(d, rule_i, cal)
        df = DateOffset(d, rule_f, cal)
        crv = GetVal(commod, "GetStrip", di, df, avg_type)
        crv = crv.corp(DateOffset(d, rule_sd, cal),
                       DateOffset(d, rule_ed, cal))
        dts += crv.dates
        vls += crv.values.tolist()

    return Curve(dts, vls).crop(start, end)



###############################################################################
#
#   Copyright: (c) 2015 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

from onyx.core import Structure, Curve, Knot, CurveIntersect
from onyx.core import GetObj, ValueType, ChildrenSet, EvalBlock

from .timeseries import get_curve_usd

import numpy as np
import math

__all__ = [
    "WithRiskValueTypes",
]


# -----------------------------------------------------------------------------
def WithRiskValueTypes(cls):
    """
    Description:
        Class decorator. It can be applied to any class implementing MktVal
        and MktValUSD Value Types.
        It adds the following Value Types to the decorated class:
            Deltas
            FxExposures
            GrossExposure
            NetExposure
            BetaAdjNetExposure
            BetaBySecurity
            ForwardVol
            VaR
            AdjVar
            MaxLoss
        These are further auxiliary Value Types used by the ones described
        above:
            RiskMktValues
    """
    # -------------------------------------------------------------------------
    def Deltas(self, graph):
        deltas = Structure()
        for kid in ChildrenSet((self, "MktValUSD"), "Spot", "Asset", graph):
            cross = "{0:3s}/USD".format(graph(kid, "Denominated"))
            spot = graph(kid, "Spot")
            fx = graph(cross, "Spot")
            shift = 0.01*spot
            with EvalBlock() as eb:
                eb.change_value(kid, "Spot", spot+shift)
                up = graph(self, "MktValUSD")
                eb.change_value(kid, "Spot", spot-shift)
                dw = graph(self, "MktValUSD")
            deltas += Structure({kid: (up-dw) / (2.0*shift*fx)})
        deltas.drop_zeros()
        return deltas

    # -------------------------------------------------------------------------
    def FxExposures(self, graph):
        exposures = Structure()
        for kid in ChildrenSet((self, "MktValUSD"),
                               "Spot", "CurrencyCross", graph):
            spot = graph(kid, "Spot")
            shift = 0.01*spot
            with EvalBlock() as eb:
                eb.change_value(kid, "Spot", spot+shift)
                up = graph(self, "MktValUSD")
                eb.change_value(kid, "Spot", spot-shift)
                dw = graph(self, "MktValUSD")
            exposures += Structure({kid: (up-dw) / (2.0*shift)})
        exposures.drop_zeros()
        return exposures

    # -------------------------------------------------------------------------
    def GrossExposure(self, graph):
        gross = 0.0
        for sec, qty in graph(self, "Deltas").items():
            spot_usd = graph(sec, "SpotUSD")
            gross += math.fabs(qty)*spot_usd
        cross = "{0:3s}/USD".format(graph(self, "Denominated"))
        fx = graph(cross, "Spot")
        return gross / fx

    # -------------------------------------------------------------------------
    def NetExposure(self, graph):
        net = 0.0
        for sec, qty in graph(self, "Deltas").items():
            spot_usd = graph(sec, "SpotUSD")
            net += qty*spot_usd
        cross = "{0:3s}/USD".format(graph(self, "Denominated"))
        fx = graph(cross, "Spot")
        return net / fx

    # -------------------------------------------------------------------------
    def BetaAdjNetExposure(self, graph):
        beta_by_sec = graph(self, "BetaBySecurity")
        beta = 0.0
        for sec, qty in graph(self, "Deltas").items():
            beta += beta_by_sec[sec]*qty*graph(sec, "SpotUSD")
        cross = "{0:3s}/USD".format(graph(self, "Denominated"))
        fx = graph(cross, "Spot")
        return beta / fx

    # -------------------------------------------------------------------------
    def BetaBySecurity(self, graph):
        # --- to improve performance:
        #     1) we fetch curves using Risk.StartDate and Risk.EndDate and then
        #        we crop them to Risk.BetaStartDate and Risk.BetaEndDate,
        #     2) we dynamically add BetaLookupTable attribute to each security
        #        and cache in there beta by (index, start date, end date).
        start = graph("Risk", "StartDate")
        end = graph("Risk", "EndDate")
        beta_start = graph("Risk", "BetaStartDate")
        beta_end = graph("Risk", "BetaEndDate")
        betas = Structure()
        index = graph("Risk", "RefIndex")
        idx_prcs_full = get_curve_usd(index, beta_start, beta_end)
        for sec in graph(self, "Deltas"):
            key = (index, beta_start, beta_end)
            sec = GetObj(sec)
            try:
                beta = sec.BetaLookupTable[key]
            except (AttributeError, KeyError) as err:
                sec_prcs = graph(sec, "PricesForRisk",
                                 start, end).crop(beta_start, beta_end)
                sec_prcs, idx_prcs = CurveIntersect([sec_prcs, idx_prcs_full])
                # --- do not attempt to calculate beta with less than 15
                #     datapoints
                if len(sec_prcs) >= 15:
                    s_rets = np.diff(np.log(sec_prcs.values))
                    s_rets -= s_rets.mean()
                    i_rets = np.diff(np.log(idx_prcs.values))
                    i_rets -= i_rets.mean()
                    beta = np.dot(i_rets, s_rets) / np.dot(i_rets, i_rets)
                else:
                    beta = None
                if isinstance(err, AttributeError):
                    # --- lookup table doesn't exist yet for this security
                    sec.BetaLookupTable = {key: beta}
                else:
                    sec.BetaLookupTable[key] = beta
            betas[sec.Name] = beta
        return betas

    # -------------------------------------------------------------------------
    def RiskMktValues(self, graph):
        start = graph("Risk", "StartDate")
        end = graph("Risk", "EndDate")

        # --- load fx to convert mkt_vls from USD to the Denominated
        #     currency
        cross = "{0:3s}/USD".format(graph(self, "Denominated"))
        fx = graph(cross, "Spot")

        mkt_vls = None
        for sec, qty in graph(self, "Deltas").items():
            prcs = graph(sec, "PricesForRisk", start, end)
            if mkt_vls is None:
                mkt_vls = prcs*qty
            else:
                prcs, mkt_vls = CurveIntersect([prcs, mkt_vls])
                mkt_vls += prcs*qty

        if mkt_vls is None:
            return Curve()
        else:
            return mkt_vls / fx

    # -------------------------------------------------------------------------
    def ForwardVol(self, graph):
        start = graph("Risk", "VaRStartDate")
        end = graph("Risk", "VaREndDate")
        mkt_values_risk = graph(self, "RiskMktValues").crop(start, end)

        return risk_metrics(mkt_values_risk)["Std"]

    # -------------------------------------------------------------------------
    def VaR(self, graph, ndays=1, pctl=95.0):
        start = graph("Risk", "VaRStartDate")
        end = graph("Risk", "VaREndDate")
        mkt_values_risk = graph(self, "RiskMktValues").crop(start, end)

        return risk_metrics(mkt_values_risk, ndays=ndays, pctl=pctl)["VaR"]

    # -------------------------------------------------------------------------
    def AdjVaR(self, graph, ndays=1, pctl=95.0, nbos_adj=60):
        start = graph("Risk", "VaRStartDate")
        end = graph("Risk", "VaREndDate")
        mkt_values_risk = graph(self, "RiskMktValues").crop(start, end)

        metrics = risk_metrics(mkt_values_risk,
                               ndays=ndays, pctl=pctl, nbos_adj=nbos_adj)

        return metrics["AdjFactor"]*["VaR"]

    # -------------------------------------------------------------------------
    def MaxLoss(self, graph, ndays=1):
        return max_loss(graph(self, "RiskMktValues"), ndays=ndays)

    # --- here we whack value types into the decorated class
    cls.Deltas = ValueType()(Deltas)
    cls.FxExposures = ValueType()(FxExposures)
    cls.GrossExposure = ValueType()(GrossExposure)
    cls.NetExposure = ValueType()(NetExposure)

    cls.BetaBySecurity = ValueType()(BetaBySecurity)
    cls.BetaAdjNetExposure = ValueType()(BetaAdjNetExposure)

    cls.RiskMktValues = ValueType()(RiskMktValues)

    cls.ForwardVol = ValueType()(ForwardVol)
    cls.VaR = ValueType("PropSubGraph")(VaR)
    cls.AdjVaR = ValueType("PropSubGraph")(AdjVaR)
    cls.MaxLoss = ValueType("PropSubGraph")(MaxLoss)

    return cls


# -----------------------------------------------------------------------------
def regress_returns(sec_prcs, idx_prcs):
    sec_prcs, idx_prcs = CurveIntersect([sec_prcs, idx_prcs])
    sec_rets = np.diff(np.log(sec_prcs.values))
    idx_rets = np.diff(np.log(idx_prcs.values))
    explanatory_vars = np.vstack([idx_rets, np.ones_like(idx_rets)]).T
    return np.linalg.lstsq(explanatory_vars, sec_rets)[0]


# -----------------------------------------------------------------------------
def risk_metrics(mkt_values, ndays=1, pctl=95.0, nbos_adj=60):
    """
    Description:
        Return key risk metrics given daily timeseries of market values.
    Inputs:
        mkt_values - a curve with daily market values for the portfolio.
        ndays      - time-horizon for calculation of risk-metrics.
        pctl       - confidence level for calculation of VaR and cVaR.
        nbos_adj   - use the last nbos_adj observations to adjust for
                     higher/lower short term volatility (for AdjVaR).
    Returns:
        A dictionary with:
            Std:       the standard deviation of the distribution of returns.
            VaR:       the Value at Risk at the desired percentile.
            cVaR:      the conditional Value at risk defined as the expected
                       value of the tail of the distribution beyhond the
                       desired percentile.
            AdjFactor: the adjustment factor to account for higher/lower short
                       term volatility.
    """
    if len(mkt_values) == 0:
        std = 0.0
        var = 0.0
        cvar = 0.0
        adj = 1.0
    else:
        # --- we flip the sign so that losses are positive numbers. we reverse
        #     the array so that the eventually unused elements (when the length
        #     of the array is not a whole multiple of ndays) are the oldest
        values = -mkt_values.values[::-1][::ndays][::-1]
        pnl_loss = np.diff(values)
        pnl_loss -= pnl_loss.mean()

        std = pnl_loss.std()
        var = np.percentile(pnl_loss, pctl)
        cvar = pnl_loss[pnl_loss >= var].mean()

        # --- the adjustment is calculated as the ratio between a short-term
        #     estimate of the standard deviation and the estimate calculated
        #     over the full range.
        adj = pnl_loss[-nbos_adj:].std() / std

    return {"Std": std, "VaR": var, "cVaR": cvar, "AdjFactor": adj}


# -----------------------------------------------------------------------------
def max_loss(mkt_values, ndays=1):
    """
    Description:
        Return the largest loss given daily timeseries of market values.
    Inputs:
        mkt_values - a curve with daily market values for the book/ portfolio.
        ndays      - time-horizon for calculation of risk-metrics.
    Returns:
        A knot with the maximum loss (expressed as a positive number) and the
        corresponding date.
    """
    if len(mkt_values) == 0:
        max_loss_val = 0.0
        max_loss_date = None
    else:
        # --- for max loss, consider all (i.e. overlapping) ndays-differences:
        #     flip the sign so that losses are positive numbers
        diffs = -(mkt_values.values[ndays:] - mkt_values.values[:-ndays])
        max_loss_idx = diffs.argmax()
        max_loss_val = diffs[max_loss_idx]

        # --- the date of max loss is meant to be the beginning of the ndays
        #     interval
        max_loss_date = mkt_values.dates[max_loss_idx].strftime("%d-%b-%Y")

    return Knot(max_loss_date, max_loss_val)



###############################################################################
#
#   Copyright: (c) 2015 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

from onyx.econometrics.gbm_functions import GbmPricePaths

import numpy as np
import math
import unittest

to_daily = 1.0 / 255.0


###############################################################################
class UnitTest(unittest.TestCase):
    # -------------------------------------------------------------------------
    def test_mean(self):
        # --- seed the random generator to make unittest deterministic
        np.random.seed(145)

        n_days, vol, rd, rf, n_sims = (255, 0.25, 0.05, 0.03, 10000)
        rd_daily = rd*to_daily
        rf_daily = rf*to_daily

        paths = GbmPricePaths(vol*math.sqrt(to_daily), n_days,
                              rd_daily, rf_daily, n_sims, False)

        exact_mean = np.exp((rd - rf)*np.cumsum(np.ones(n_days))*to_daily)
        error = np.abs(paths[1:,:].mean(axis=1) / exact_mean - 1.0)

        # --- convergence within 1%
        self.assertFalse(np.any(error > 0.01))

    # -------------------------------------------------------------------------
    def test_std(self):
        # --- seed the random generator to make unittest deterministic
        np.random.seed(145)

        n_days, vol, rd, rf, n_sims = (255, 0.25, 0.05, 0.03, 100000)
        rd_daily = rd*to_daily
        rf_daily = rf*to_daily

        paths = GbmPricePaths(vol*math.sqrt(to_daily), n_days,
                              rd_daily, rf_daily, n_sims, False)

        exact_std = vol*np.sqrt(np.cumsum(np.ones(n_days))*to_daily)
        error = np.abs(np.log(paths[1:,:]).std(axis=1) / exact_std - 1.0)

        # --- convergence within 1%
        self.assertFalse(np.any(error > 0.01))


if __name__ == "__main__":
    from onyx.core.utils.unittest import UseEphemeralDbs
    with UseEphemeralDbs():
        unittest.main(failfast=True)



###############################################################################
#
#   Copyright: (c) 2015 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

from onyx.core import ReferenceField, FloatField
from onyx.core import GetVal, GraphNodeVt, RetainedFactory
from onyx.core import DiscardInheritedAttribute, InheritAsProperty

from .ufo_equity_asset import EquityAsset

__all__ = ["DepositaryReceipt"]

DISCARD = ["RiskProxy"]
REPLACE = ["Country", "Region", "Sector", "Subsector"]


###############################################################################
@InheritAsProperty(REPLACE, "Parent")
@DiscardInheritedAttribute(DISCARD)
class DepositaryReceipt(EquityAsset):
    """
    class used to represent a depositary receipt (GDR, ADR, etc).
    """
    # --- a reference to the local listing for this company
    Parent = ReferenceField(obj_type="EquityAsset")
    # --- one depositary receipt corresponds to these many shares in the
    #     local listing
    Conversion = FloatField(default=1.0)

    # -------------------------------------------------------------------------
    def __post_init__(self):
        # --- here we don't want to call EquityAsset's __post_init__ method
        #     but rather that of its super-class: this to avoid trying to set
        #     properties that have been inherited as properties.
        super(EquityAsset, self).__post_init__()

        exchange_code = GetVal(self.Exchange, "Code")

        self.Name = self.name_fmt.format(self.Symbol, exchange_code)

        # --- name of the time-series that store historical adjusted prices,
        #     historical closes (marks) and dividends
        fields = self.Symbol, exchange_code
        self.TimeSeries = "EQ-TS {0:s} {1:s}".format(*fields)
        self.Marks = "EQ-MKS {0:s} {1:s}".format(*fields)
        self.Dividends = "EQ-DIV {0:s} {1:s}".format(*fields)

    # -------------------------------------------------------------------------
    @GraphNodeVt("Property")
    def RiskProxy(self, graph):
        return graph(self, "Parent")

    # -------------------------------------------------------------------------
    @RetainedFactory()
    def Spot(self, graph):
        """
        Return the official close value as of MktDataDate (or the most recent
        close if ForceStrict is False) in the Denominated currency.
        """
        fx = graph(self, "DenominatedToLocal")
        return self.spot_local_ccy() * graph(self, "Multiplier") / fx

    # -------------------------------------------------------------------------
    @GraphNodeVt("Property")
    def DenominatedToLocal(self, graph):
        parent = graph(self, "Parent")
        den_to_usd = "{0:3s}/USD".format(GetVal(self, "Denominated"))
        loc_to_usd = "{0:3s}/USD".format(GetVal(parent, "Denominated"))
        return graph(den_to_usd, "Spot") / graph(loc_to_usd, "Spot")

    # -------------------------------------------------------------------------
    def spot_local_ccy(self):
        """
        Return (off-graph) spot of the depositary receipt converted to the
        currency of the local listing.
        This gymnastic is needed to reflect the proper FX risk.
        """
        spot = GetVal(GetVal(self, "Marks"), "Price")
        return spot * GetVal(self, "DenominatedToLocal")



###############################################################################
#
#   Copyright: (c) 2015 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

from onyx.core import Date, Knot, Curve, HlocvCurve, LYY2Date, DateOffset
from onyx.core import DelObj, UpdateObj, ObjNotFound
from onyx.core import GetVal, SetVal, ValueType
from onyx.core import DictField, StringField, ReferenceField
from onyx.core import InheritAsProperty, RetainedFactory

from ..system.ufo_asset import Asset

__all__ = ["EquityIndexCnt"]

# --- replace all base class stored attributes with pointers to EquityIndex,
#     with the exception of RiskProxy
REPLACE = Asset._json_fields.difference({"RiskProxy"})
# --- add a few EquityIndex-specific stored attributes
REPLACE = REPLACE.union({"Country", "Region",
                         "Sector", "Subsector", "ContractSize"})


###############################################################################
@InheritAsProperty(REPLACE, "EquityIndex")
class EquityIndexCnt(Asset):
    """
    This class used to access equity-index contract information and the
    relative price marks.
    """
    # --- this is the parent object that babysits all contracts on the same
    #     commodity
    EquityIndex = ReferenceField(obj_type="EquityIndex")

    # --- DeliveryMonth is the LYY code for the contract
    DeliveryMonth = StringField()

    Tickers = DictField()
    Marks = StringField()

    # -------------------------------------------------------------------------
    def __post_init__(self):
        sym = GetVal(self.EquityIndex, "Symbol")
        args = sym, self.DeliveryMonth

        self.Name = self.get_name(*args)
        self.Marks = "CNT-MKS {0:s} {1:3s}".format(*args)

    # -------------------------------------------------------------------------
    @ValueType()
    def Ticker(self, graph):
        return graph(self, "Tickers")[graph("Settings", "Platform")]

    # -------------------------------------------------------------------------
    @ValueType()
    def UniqueId(self, graph):
        sym = graph(self, "Symbol")
        mth = graph(self, "DeliveryMonth")
        return "{0:s} {1:3s}".format(sym, mth)

    # -------------------------------------------------------------------------
    @ValueType()
    def FutSettDate(self, graph):
        rule = graph(self, "SettDateRule")
        cal = graph(self, "HolidayCalendar")
        mth = graph(self, "DeliveryMonth")
        return DateOffset(LYY2Date(mth), rule, cal)

    # -------------------------------------------------------------------------
    @ValueType()
    def VolEndDate(self, graph):
        return graph(self, "FutSettDate")

    # -------------------------------------------------------------------------
    @ValueType()
    def OptExpDate(self, graph):
        rule = graph(self, "OptExpDateRule")
        cal = graph(self, "HolidayCalendar")
        return DateOffset(graph(self, "FutSettDate"), rule, cal)

    # -------------------------------------------------------------------------
    @RetainedFactory()
    def Spot(self, graph):
        """
        Return the official close value as of MktDataDate (or the most recent
        close if ForceStrict is False) in the Denominated currency.
        """
        return graph(graph(self, "Marks"), "Price")*graph(self, "Multiplier")

    # -------------------------------------------------------------------------
    @ValueType()
    def Last(self, graph):
        """
        Return the knot with the most recent close value (irrespective of
        MktDataDate) in the Denominated currency.
        """
        marks = graph(self, "Marks")
        date, value = graph(marks, "LastKnot", date=Date.high_date())
        return Knot(date, value*graph(self, "Multiplier"))

    # -------------------------------------------------------------------------
    @ValueType("PropSubGraph")
    def GetMarks(self, graph, start=None, end=None):
        return graph(graph(self, "Marks"), "PrcFixCurve", start, end)

    # -------------------------------------------------------------------------
    @ValueType("PropSubGraph")
    def GetCurve(self, graph, start=None, end=None, field=None):
        if field == "Close":
            return graph(self, "GetMarks", start, end)
        else:
            return HlocvCurve() if field is None else Curve()

    # -------------------------------------------------------------------------
    def delete(self):
        # --- delete price fixes
        try:
            DelObj(self.Marks)
        except ObjNotFound:
            pass

        # --- remove from set of contracts for the CommodAsset object
        cnts = GetVal(self.EquityIndex, "Contracts")
        cnts.discard(self.DeliveryMonth)

        SetVal(self.EquityIndex, "Contracts", cnts)
        UpdateObj(self.EquityIndex)

    # -------------------------------------------------------------------------
    @classmethod
    def get_name(cls, symbol, del_mth):
        """
        Generate contract's name from Symbol, and DeliveryMonth
        """
        return "CNT {0:s} {1:3s}".format(symbol, del_mth)


# -----------------------------------------------------------------------------
def prepare_for_test():
    from onyx.core import Date, RDate, Date2LYY, AddIfMissing, EvalBlock

    from ..system.ufo_price_fix import PriceFix
    from ..system import ufo_price_fix
    from . import ufo_equity_index

    ufo_equity_index.prepare_for_test()
    ufo_price_fix.prepare_for_test()

    mth = Date2LYY(Date.today() + RDate("+E"))
    bbg_ticker = "VG{0:s}{1:s}".format(mth[0], mth[-1])

    sx5e_info = {
        "EquityIndex": "EQ-IDX SX5E",
        "DeliveryMonth": mth,
        "Tickers": {"Bloomberg": bbg_ticker},
    }

    cnt = AddIfMissing(EquityIndexCnt(**sx5e_info))
    prc = 3000.0

    prc_fix = AddIfMissing(PriceFix(Name=GetVal(cnt, "Marks")))
    with EvalBlock() as eb:
        eb.change_value("Database", "ArchivedOverwritable", True)
        prc_fix.set_dated("Price", Date.today(), prc)

    return [(cnt.Name, prc)]



###############################################################################
#
#   Copyright: (c) 2015 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

from onyx.core import Date, Knot, GetVal, OnyxTestCase
from onyx.system import ufo_currency_cross

import unittest


###############################################################################
class UnitTest(OnyxTestCase):
    # -------------------------------------------------------------------------
    def setUp(self):
        super().setUp()
        self.marks = ufo_currency_cross.prepare_for_test()

    # -------------------------------------------------------------------------
    def test_Spot(self):
        for cross, value in self.marks.items():
            self.assertEqual(GetVal(cross, "Spot"), value)

    # -------------------------------------------------------------------------
    def test_Last(self):
        for cross, value in self.marks.items():
            self.assertIsInstance(GetVal(cross, "Last"), Knot)
            self.assertEqual(GetVal(cross, "Last"),
                             Knot(Date.today(), value))

    # -------------------------------------------------------------------------
    def test_GetCurve(self):
        for cross, value in self.marks.items():
            crv = GetVal(cross, "GetCurve")
            self.assertTrue(len(crv) > 0)
            self.assertEqual(crv.back.date, Date.today())
            self.assertEqual(crv.back.value, value)


if __name__ == "__main__":
    from onyx.core import UseEphemeralDbs
    with UseEphemeralDbs():
        unittest.main(failfast=True)



###############################################################################
#
#   Copyright: (c) 2015 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

from onyx.core import GetVal, UfoBase, ValueType
from onyx.core import FloatField, ReferenceField, StringField

from ..risk.timeseries import prices_for_risk

import abc

__all__ = ["WithAssetAttributes", "Asset"]


###############################################################################
class WithAssetAttributes(UfoBase):
    """
    This class is used to expose stored attributes that are common to all
    assets or to classes that act as parents for assets (such as CommodAsset).
    """
    Symbol = StringField()
    Exchange = ReferenceField(obj_type="Exchange")
    Denominated = ReferenceField(obj_type="Currency")
    HolidayCalendar = ReferenceField(obj_type="HolidayCalendar")
    Multiplier = FloatField(default=1.0)
    Description = StringField(default="")
    RiskProxy = ReferenceField(obj_type="WithAssetAttributes")
    VolMatrix = StringField()
    VolModel = StringField()

    # -------------------------------------------------------------------------
    def __post_init__(self):
        # --- set defaults for attributes inherited from Exchange
        for attr in {"Denominated", "HolidayCalendar", "Multiplier"}:
            if getattr(self, attr) is None:
                setattr(self, attr, GetVal(self.Exchange, attr))


###############################################################################
class Asset(WithAssetAttributes):
    """
    Base class used to provide a generic interface to all assets.
    """
    # -------------------------------------------------------------------------
    @abc.abstractmethod
    def UniqueId(self, graph):
        raise NotImplementedError()

    # -------------------------------------------------------------------------
    # this should always be implemented as a Settable Property using the
    # RetainedFactory
    @abc.abstractmethod
    def Spot(self, graph):
        raise NotImplementedError()

    # -------------------------------------------------------------------------
    @abc.abstractmethod
    def GetCurve(self, graph, start, end, field):
        raise NotImplementedError()

    # -------------------------------------------------------------------------
    @ValueType()
    def SpotUSD(self, graph):
        cross = "{0:3s}/USD".format(graph(self, "Denominated"))
        return graph(self, "Spot")*graph(cross, "Spot")

    # -------------------------------------------------------------------------
    @ValueType("Callable")
    def PricesForRisk(self, graph, start, end):
        return prices_for_risk(graph(self, "Name"), start, end, strict=False)



###############################################################################
#
#   Copyright: (c) 2015 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

from onyx.core import GCurve, Interpolate, RDate, DateRange, Structure
from onyx.core import GetObj, AddObj, UpdateObj, ExistsInDatabase
from onyx.core import ObjDbTransaction, ObjNotFound, ObjDbQuery
from onyx.core import EvalBlock, GraphScope, GetVal, SetVal

from ..system.ufo_portfolio import Portfolio

__all__ = [
    "get_historical",
    "pnl_by_long_short",
    "update_attribution_hierarchy",
    "stats_by_portfolio",
    "stats_by_container",
]

# --- query to get all assets traded in a given book, skipping all trades that
#     have been subsequently deleted or moved to a different book
ASSETS_BY_BOOK = """
SELECT DISTINCT(Objects.Data->>'Asset') AS Asset FROM (
    SELECT DISTINCT(Objects.Data->>'SecurityTraded') AS SecTraded FROM (
        SELECT Trade As TradeName FROM (
            SELECT Trade, SUM(Qty) AS TotQty
            FROM PosEffects WHERE Book=%s AND UnitType<>'ForwardCash'
            GROUP BY Trade)
        AS Pivot WHERE Pivot.TotQty<>0) AS Trades
    INNER JOIN Objects ON Objects.Name=Trades.TradeName) AS Tradables
INNER JOIN Objects ON Objects.Name=Tradables.SecTraded AND
                      Objects.Data->>'Asset'<>'';
"""


# -----------------------------------------------------------------------------
def get_historical(port, start, end, fields, fund=None):
    """
    Description:
        Get historical values for a given portfolio.
    Inputs:
        port   - portfolio name
        start  - start date
        end    - end date
        fields - a list of fields
        fund   - this field must specify the fund's name whenever aum is one
                 of the required fields
    Returns:
        A GCurve of dictionaries with field values.
    """
    fields = dict.fromkeys(fields, True)

    add_mktval = fields.pop("mktval", False)
    add_gross = fields.pop("gross", False)
    add_net = fields.pop("net", False)
    add_fvol = fields.pop("fvol", False)
    add_var = fields.pop("var", False)
    add_aum = fields.pop("aum", False)
    add_nav = fields.pop("nav", False)

    if len(fields):
        raise ValueError("Unrecognized fields: {0!s}".format(fields))

    results = GCurve()
    for date in DateRange(start, end, "+1b"):
        values = {}
        with EvalBlock() as eb:
            # --- this shifts back both market and positions
            eb.change_value("Database", "PricingDate", date)
            if add_mktval:
                values["mktval"] = GetVal(port, "MktValUSD")
            if add_gross:
                values["gross"] = GetVal(port, "GrossExposure")
            if add_net:
                values["net"] = GetVal(port, "NetExposure")
            if add_fvol:
                values["fvol"] = GetVal(port, "ForwardVol")
            if add_var:
                values["var"] = GetVal(port, "VaR")
            if add_aum and fund is not None:
                values["aum"] = GetVal(fund, "Aum")
            if add_nav and fund is not None:
                values["nav"] = GetVal(fund, "Nav")

        results[date] = values

    return results


# -----------------------------------------------------------------------------
def pnl_by_long_short(port, start, end):
    """
    Description:
        Return historical daily P&L for long and short positions of a given
        portfolio, using the delta approximation.
        NB: We use adjusted prices to include the effect of dividends and
            assume that all positions are finaced in local currency (so that
            there is FX risk on the daily P&L only).
    Inputs:
        port  - portfolio name
        start - start date
        end   - end date
    Returns:
        A GCurve of dictionaries with field values.
    """
    crv_start = start + RDate("-1w")

    def get_price(sec, date):
        crv = GetVal(sec, "GetCurve", crv_start, end, field="Close")
        mul = GetVal(sec, "Multiplier")
        return mul*Interpolate(crv, date)

    def get_fx_rate(sec, date):
        cross = "{0:3s}/USD".format(GetVal(sec, "Denominated"))
        crv = GetVal(cross, "GetCurve", crv_start, end)
        return Interpolate(crv, date)

    def get_pos_and_prices(port, date):
        with EvalBlock() as eb:
            eb.change_value("Database", "PricingDate", date)
            pos = GetVal(port, "Deltas")
        return pos, {sec: get_price(sec, date) for sec in pos}

    date = start + RDate("-1b")
    old_pos, old_prcs = get_pos_and_prices(port, date)
    results = GCurve()

    for date in DateRange(start, end, "+1b"):
        fx_rates = {sec: get_fx_rate(sec, date) for sec in old_pos}
        long = short = 0.0
        for sec, qty in old_pos.items():
            pnl = qty*(get_price(sec, date) - old_prcs[sec])*fx_rates[sec]
            if qty >= 0.0:
                long += pnl
            else:
                short += pnl

        results[date] = {"long": long, "short": short}

        old_pos, old_prcs = get_pos_and_prices(port, date)

    return results


# -----------------------------------------------------------------------------
def update_attribution_hierarchy(fund):
    """
    Description:
        Updates the equity attribution hierarchy for all books that are leaves
        of a given fund.
    Inputs:
        fund - the fund's name
    Returns:
        None
    """
    port = GetVal(fund, "Portfolio")
    books = GetVal(port, "Books")
    fund_ccy = GetVal(fund, "Denominated")

    with ObjDbTransaction("Rebuild hierarchy"):
        # --- first wipe clean all attribution portfolios
        for container in ("Countries", "Regions", "Sectors", "Subsectors"):
            for bucket in GetVal(container, "Items"):
                name = "{0:s} {1:s}".format(port, bucket.upper())
                try:
                    SetVal(name, "Children", Structure())
                except ObjNotFound:
                    continue

        # --- then add each book in the hierarchy to the relevant attribution
        #     portfolios
        for book in books:
            # --- lookup the set of all assets traded in this book
            assets = {rec.asset for rec in
                      ObjDbQuery(ASSETS_BY_BOOK, (book, ), attr="fetchall")}
            if not len(assets):
                continue

            for attr in ("Country", "Region", "Sector", "Subsector"):
                buckets = {GetVal(asset, attr) for asset in assets}
                qty = 1.0 / float(len(buckets))

                for bucket_name in buckets:
                    name = "{0:s} {1:s}".format(port, bucket_name.upper())
                    try:
                        bucket = GetObj(name)
                    except ObjNotFound:
                        bucket = AddObj(Portfolio(Name=name,
                                                  DisplayName=bucket_name,
                                                  Denominated=fund_ccy))
                    children = bucket.Children or Structure()
                    children[book] = qty
                    bucket.Children = children
                    UpdateObj(bucket)


# -----------------------------------------------------------------------------
def stats_by_portfolio(fund, month):
    """
    Description:
        Return attribution statistics (Gross Exposure, Net Exposure, and MTD
        P&L) on a given month for all the children of a given fund.
    Inputs:
        fund  - the fund's name
        month - the month's date (usually the last day)
    Returns:
        A dictionary.
    """
    port = GetVal(fund, "Portfolio")
    start = month + RDate("-1m+e")
    end = month + RDate("+e")
    stats = {}

    scope_start = GraphScope()
    scope_start.change_value("Database", "PricingDate", start)
    scope_end = GraphScope()
    scope_end.change_value("Database", "PricingDate", end)

    with scope_end:
        stats = {
            "date": end,
            "aum": GetVal(fund, "Aum"),
            "nav": GetVal(fund, "Nav"),
        }

    for kid in GetVal(port, "Children"):
        with scope_start:
            mktval_start = GetVal(kid, "MktValUSD")
        with scope_end:
            info = {
                "gross": GetVal(kid, "GrossExposure"),
                "net": GetVal(kid, "NetExposure"),
                "mtd pnl": GetVal(kid, "MktValUSD") - mktval_start,
            }
        stats[GetVal(kid, "DisplayName")] = info

    return stats


# -----------------------------------------------------------------------------
def stats_by_container(container, fund, month):
    """
    Description:
        Return attribution statistics (Gross Exposure, Net Exposure, and MTD
        P&L) on a given month for all the portfolios in a given container.
    Inputs:
        container - the container's name
        fund      - the fund's name (used to get AUM)
        month     - the month's date (usually the last day)
    Returns:
        A dictionary.
    """
    port = GetVal(fund, "Portfolio")
    start = month + RDate("-1m+e")
    end = month + RDate("+e")
    stats = {}

    scope_start = GraphScope()
    scope_start.change_value("Database", "PricingDate", start)
    scope_end = GraphScope()
    scope_end.change_value("Database", "PricingDate", end)

    with scope_end:
        stats = {
            "date": end,
            "aum": GetVal(fund, "Aum"),
            "nav": GetVal(fund, "Nav"),
        }

    for item in GetVal(container, "Items"):
        kid = "{0:s} {1:s}".format(port, item.upper())
        if not ExistsInDatabase(kid):
            continue
        with scope_start:
            mktval_start = GetVal(kid, "MktValUSD")
        with scope_end:
            info = {
                "gross": GetVal(kid, "GrossExposure"),
                "net": GetVal(kid, "NetExposure"),
                "mtd pnl": GetVal(kid, "MktValUSD") - mktval_start,
            }
        stats[GetVal(kid, "DisplayName")] = info

    return stats



###############################################################################
#
#   Copyright: (c) 2015 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

from onyx.core import ValueType, ReferenceField

from .tradable_api import NamedByInference

__all__ = ["CfdFixedLeg"]


###############################################################################
class CfdFixedLeg(NamedByInference):
    """
    Fixed leg of a contract for difference (equity swap, futures, etc).
    """
    Currency = ReferenceField(obj_type="Currency")

    # -------------------------------------------------------------------------
    @ValueType()
    def MktVal(self, graph):
        return 1.0

    # -------------------------------------------------------------------------
    @ValueType()
    def MktValUSD(self, graph):
        return graph("{0:3s}/USD".format(graph(self, "Currency")), "Spot")

    # -------------------------------------------------------------------------
    @property
    def ImpliedName(self):
        return "CfdFixed {0:3s}".format(self.Currency)



###############################################################################
#
#   Copyright: (c) 2015 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

from onyx.core import Date, LYY2Date, Curve, HlocvCurve, Knot, DateOffset
from onyx.core import DelObj, UpdateObj, ObjNotFound
from onyx.core import TsDbGetCurve, TsDbQuery, TsNotFound
from onyx.core import GetVal, SetVal, ValueType
from onyx.core import StringField, DictField, ReferenceField
from onyx.core import InheritAsProperty, RetainedFactory

from ..system.ufo_asset import Asset

__all__ = ["CommodCnt"]

# --- replace all base class stored attributes with pointers to CommodAsset,
#     with the exception of RiskProxy
REPLACE = Asset._json_fields.difference({"RiskProxy"})
# --- add a few CommodAsset-specific stored attributes
REPLACE = REPLACE.union({"Market", "CntType", "ContractSize", "SettDateRule",
                         "OptExpDateRule", "Country", "Region", "Sector",
                         "Subsector"})


###############################################################################
@InheritAsProperty(REPLACE, "CommodAsset")
class CommodCnt(Asset):
    """
    This class used to access commod contract information and the relative
    timeseries.
    """
    # --- this is the parent object that babysits all contracts on the same
    #     commodity
    CommodAsset = ReferenceField(obj_type="CommodAsset")
    Tickers = DictField()

    # --- DeliveryMonth is the LYY code for the contract
    DeliveryMonth = StringField()

    Marks = StringField()
    TimeSeries = StringField()

    # -------------------------------------------------------------------------
    def __post_init__(self):
        mkt = GetVal(self.CommodAsset, "Market")
        sym = GetVal(self.CommodAsset, "Symbol")
        args = mkt, sym, self.DeliveryMonth

        self.Name = self.get_name(*args)
        self.Marks = "CNT-MKS {0:s} {1:s} {2:3s}".format(*args)
        self.TimeSeries = "CNT-TS {0:s} {1:s} {2:3s}".format(*args)

    # -------------------------------------------------------------------------
    @ValueType()
    def Ticker(self, graph):
        return graph(self, "Tickers")[graph("Settings", "Platform")]

    # -------------------------------------------------------------------------
    @ValueType()
    def UniqueId(self, graph):
        mkt = graph(self, "Market")
        sym = graph(self, "Symbol")
        mth = graph(self, "DeliveryMonth")
        return "{0:s} {1:s} {2:3s}".format(mkt, sym, mth)

    # -------------------------------------------------------------------------
    @ValueType()
    def FutSettDate(self, graph):
        rule = graph(self, "SettDateRule")
        cal = graph(self, "HolidayCalendar")
        mth = graph(self, "DeliveryMonth")
        return DateOffset(LYY2Date(mth), rule, cal)

    # -------------------------------------------------------------------------
    @ValueType()
    def VolEndDate(self, graph):
        return graph(self, "FutSettDate")

    # -------------------------------------------------------------------------
    @ValueType()
    def OptExpDate(self, graph):
        rule = graph(self, "OptExpDateRule")
        cal = graph(self, "HolidayCalendar")
        return DateOffset(graph(self, "FutSettDate"), rule, cal)

    # -------------------------------------------------------------------------
    @RetainedFactory()
    def Spot(self, graph):
        """
        Return the official close value as of MktDataDate (or the most recent
        close if ForceStrict is False) in the Denominated currency.
        """
        return graph(graph(self, "Marks"), "Price")*graph(self, "Multiplier")

    # -------------------------------------------------------------------------
    @ValueType()
    def Last(self, graph):
        """
        Return the knot with the most recent close value (irrespective of
        MktDataDate) in the Denominated currency.
        """
        marks = graph(self, "Marks")
        date, value = graph(marks, "LastKnot", date=Date.high_date())
        return Knot(date, value*graph(self, "Multiplier"))

    # -------------------------------------------------------------------------
    @ValueType("PropSubGraph")
    def GetCurve(self, graph, start=None, end=None, field=None):
        name = graph(self, "TimeSeries")
        if graph(self, "CntType") == "HLOCV":
            try:
                return TsDbGetCurve(name, start, end, "HLOCV", field)
            except TsNotFound:
                return HlocvCurve() if field is None else Curve()
        else:
            try:
                return TsDbGetCurve(name, start, end, "CRV", field)
            except TsNotFound:
                return Curve()

    # -------------------------------------------------------------------------
    @ValueType("PropSubGraph")
    def GetMarks(self, graph, start=None, end=None):
        return graph(graph(self, "Marks"), "PrcFixCurve", start, end)

    # -------------------------------------------------------------------------
    def delete(self):
        # --- delete time-series from TsDb
        if self.CntType == "HLOCV":
            query = "DELETE FROM HlocvCurves WHERE Name=%s"
        else:
            query = "DELETE FROM Curves WHERE Name=%s"

        # --- delete timeseries
        TsDbQuery(query, parms=(self.TimeSeries,))

        # --- delete price fixes
        try:
            DelObj(self.Marks)
        except ObjNotFound:
            pass

        # --- remove from set of contracts for the CommodAsset object
        cnts = GetVal(self.CommodAsset, "Contracts")
        cnts.discard(self.DeliveryMonth)

        SetVal(self.CommodAsset, "Contracts", cnts)
        UpdateObj(self.CommodAsset)

    # -------------------------------------------------------------------------
    @classmethod
    def get_name(cls, market, symbol, del_mth):
        """
        Generate contract's name from Market, Symbol, and DeliveryMonth
        """
        return "CNT {0:s} {1:s} {2:3s}".format(market, symbol, del_mth)


# -----------------------------------------------------------------------------
def prepare_for_test():
    from onyx.core import Date, RDate, Date2LYY
    from onyx.core import GetObj, AddIfMissing, EvalBlock
    from . import ufo_commod_asset
    from ..system import ufo_price_fix

    ufo_commod_asset.prepare_for_test()
    ufo_price_fix.prepare_for_test()

    mth = Date2LYY(Date.today() + RDate("+E"))
    bbg_ticker = "MO{0:s}{1:s}".format(mth[0], mth[-1])

    commod = GetObj("COMMOD CO2 EUA")
    cnt = commod.add_contract(mth,  {"Bloomberg": bbg_ticker})
    prc = 7.0

    prc_fix = AddIfMissing(ufo_price_fix.PriceFix(Name=GetVal(cnt, "Marks")))
    with EvalBlock() as eb:
        eb.change_value("Database", "ArchivedOverwritable", True)
        prc_fix.set_dated("Price", Date.today(), prc)

    return [(cnt, prc)]



###############################################################################
#
#   Copyright: (c) 2015 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

from onyx.core import UfoBase, ReferenceField, BoolField

__all__ = ["Trader"]


###############################################################################
class Trader(UfoBase):
    """
    Abacus class used to represent a system user (generally a trader).
    """
    Group = ReferenceField(obj_type="Group")
    Superuser = BoolField(default=False)


# -----------------------------------------------------------------------------
def prepare_for_test():
    from onyx.core import AddIfMissing
    from . import ufo_group

    groups = ufo_group.prepare_for_test()

    traders = [
        AddIfMissing(Trader(Name="TEST_TRADER1", Group=groups[0])),
        AddIfMissing(Trader(Name="TEST_TRADER2", Group=groups[0])),
    ]
    return [trader.Name for trader in traders]



###############################################################################
#
#   Copyright: (c) 2015 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

from onyx.core import Structure, DateOffset, ValueType, GetVal, GraphScope
from onyx.core import FloatField, StringField, ReferenceField

from ..system.tradable_api import AgingTradableObj
from ..system.tradable_api import AddByInference, HashStoredAttrs
from ..system.ufo_cfd_fixed_leg import CfdFixedLeg
from ..system.ufo_cfd_floating_leg import CfdFloatingLeg
from ..system.ufo_forward_cash import ForwardCash

__all__ = ["EquityIndexFutures"]


###############################################################################
class EquityIndexFutures(AgingTradableObj):
    """
    Tradable class that represents a equity index futures contract.
    """
    Asset = ReferenceField(obj_type="EquityIndexCnt")
    FixedPrice = FloatField(positive=True)

    # --- the settlement currency can differ from the denomination of the asset
    SettlementCcy = ReferenceField(obj_type="Currency")

    # --- fx rate uesd to convert the fixed price (which is expressed in the
    #     currency of the asset) to the settlement currency
    SettlementFx = FloatField()

    # --- the date rule for the cash payment, defaults to 2 biz days after
    #     settlement
    PaymentDateRule = StringField("+2b")

    # -------------------------------------------------------------------------
    def __post_init__(self):
        asset_ccy = GetVal(self.Asset, "Denominated")

        self.SettlementCcy = self.SettlementCcy or asset_ccy

        if self.SettlementFx is None:
            # --- default to spot
            if asset_ccy == self.SettlementCcy:
                fx = 1.0
            else:
                asset_cross = "{0:3s}/USD".format(asset_ccy)
                sett_cross = "{0:3s}/USD".format(self.SettlementCcy)
                fx = GetVal(asset_cross, "Spot") / GetVal(sett_cross, "Spot")

            self.SettlementFx = fx

    # -------------------------------------------------------------------------
    @ValueType()
    def Leaves(self, graph):
        asset = graph(self, "Asset")
        prc = graph(self, "FixedPrice")
        mul = graph(asset, "Multiplier")
        lot = graph(asset, "ContractSize")

        sett_ccy = graph(self, "SettlementCcy")
        sett_fx = graph(self, "SettlementFx")

        # --- leaves are created in memory
        floating = AddByInference(CfdFloatingLeg(Asset=asset), True)
        fixed = AddByInference(CfdFixedLeg(Currency=sett_ccy), True)

        return Structure({
            floating.Name: lot,
            fixed.Name: -lot*prc*mul*sett_fx,
        })

    # -------------------------------------------------------------------------
    @ValueType()
    def MktValUSD(self, graph):
        mtm = 0.0
        for sec, qty in graph(self, "Leaves").items():
            mtm += qty*graph(sec, "MktValUSD")
        return mtm

    # -------------------------------------------------------------------------
    @ValueType()
    def MktVal(self, graph):
        sett_ccy = graph(self, "SettlementCcy")
        spot_fx = graph("{0:3s}/USD".format(sett_ccy), "Spot")
        return graph(self, "MktValUSD") / spot_fx

    # -------------------------------------------------------------------------
    @ValueType()
    def NextTransactionDate(self, graph):
        return graph(graph(self, "Asset"), "FutSettDate")

    # -------------------------------------------------------------------------
    @ValueType()
    def TradeTypes(self, graph):
        mapping = super().TradeTypes
        mapping.update({
            "Settlement": "SettlementSecurities",
        })
        return mapping

    # -------------------------------------------------------------------------
    @ValueType()
    def ExpectedTransaction(self, graph):
        return "Settlement"

    # -------------------------------------------------------------------------
    @ValueType()
    def SettlementSecurities(self, graph):
        asset = graph(self, "Asset")
        sett_ccy = graph(self, "SettlementCcy")
        sett_date = graph(asset, "FutSettDate")
        pay_rule = graph(self, "PaymentDateRule")
        cal = graph(asset, "HolidayCalendar")

        cash = ForwardCash(Currency=sett_ccy,
                           PaymentDate=DateOffset(sett_date, pay_rule, cal))

        # --- the market value at the settlement date corresponds to the
        #     amount of cash that needs to exchange hands
        with GraphScope() as scope:
            scope.change_value("Database", "MktDataDate", sett_date)
            pay_qty = graph(self, "MktVal")

        return [
            {"Security": AddByInference(cash, in_memory=True),
             "Quantity": pay_qty},
        ]

    # -------------------------------------------------------------------------
    @property
    def ImpliedName(self):
        sym = GetVal(self.Asset, "Symbol")
        mth = GetVal(self.Asset, "DeliveryMonth")
        mush = HashStoredAttrs(self, 4)
        return ("EqIdxFut {0:s} {1:3s} {2:3s} {3:4s} "
                "{{0:2d}}").format(sym, mth, self.SettlementCcy, mush)


# -----------------------------------------------------------------------------
def prepare_for_test():
    from . import ufo_equity_index_contract
    from ..system import ufo_forward_cash

    cnt, prc = ufo_equity_index_contract.prepare_for_test()[0]
    ufo_forward_cash.prepare_for_test()

    info = {
        "Asset": cnt,
        "FixedPrice": prc - 100.0,
    }

    securities = [AddByInference(EquityIndexFutures(**info))]

    return [sec.Name for sec in securities]



###############################################################################
#
#   Copyright: (c) 2015 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

from onyx.core import GetVal, GetObj, Structure, ChildrenSet, OnyxTestCase
from onyx.system import ufo_cash_balance

import unittest


###############################################################################
class UnitTest(OnyxTestCase):
    # -------------------------------------------------------------------------
    def setUp(self):
        super().setUp()
        self.securities = ufo_cash_balance.prepare_for_test()
        assert "CASH USD" in self.securities
        assert "CASH EUR" in self.securities
        assert "CASH GBP" in self.securities

    # -------------------------------------------------------------------------
    def test_Leaves(self):
        self.assertEqual(GetVal("CASH USD", "Leaves"),
                         Structure([("CASH USD", 1.0)]))

    # -------------------------------------------------------------------------
    def test_MktVal(self):
        self.assertEqual(GetVal("CASH USD", "MktVal"), 1.0)
        self.assertEqual(GetVal("CASH EUR", "MktVal"), 1.0)
        self.assertEqual(GetVal("CASH GBP", "MktVal"), 1.0)

    # -------------------------------------------------------------------------
    def test_MktValUSD(self):
        self.assertEqual(GetVal("CASH USD", "MktValUSD"), 1.00)
        self.assertEqual(GetVal("CASH EUR", "MktValUSD"), 1.15)
        self.assertEqual(GetVal("CASH GBP", "MktValUSD"), 1.50)

    # -------------------------------------------------------------------------
    def test_ExpectedSecurities(self):
        sec = GetObj("CASH USD")
        self.assertEqual(GetVal("CASH USD", "ExpectedSecurities", "Buy"),
                         [{"Security": sec, "Quantity": 1.0}])
        self.assertEqual(GetVal("CASH USD", "ExpectedSecurities", "Sell"),
                         [{"Security": sec, "Quantity": -1.0}])

    # -------------------------------------------------------------------------
    def test_Children(self):
        kids = ChildrenSet(("CASH USD", "MktValUSD"), "Spot")
        self.assertEqual(kids, {"USD/USD"})
        kids = ChildrenSet(("CASH EUR", "MktValUSD"), "Spot")
        self.assertEqual(kids, {"EUR/USD"})
        kids = ChildrenSet(("CASH GBP", "MktValUSD"), "Spot")
        self.assertEqual(kids, {"GBP/USD"})


if __name__ == "__main__":
    from onyx.core import UseEphemeralDbs
    with UseEphemeralDbs():
        unittest.main(failfast=True)



###############################################################################
#
#   Copyright: (c) 2015 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

from onyx.quantlib.black_scholes import opt_BS, opt_BAW, opt_JZ, implied_vol
from onyx.quantlib.tree_pricers import opt_BT

import unittest

# --- reference table of black-scholes results (from gnumeric)
BLACK_SCHOLES_REF = {
    # --- short dated, negative carry
    ("C", 40.0, 50.0, 0.20, 0.25, 0.025, 0.050): (0.016496, 0.012264),
    ("P", 40.0, 50.0, 0.20, 0.25, 0.025, 0.050): (10.201858, -0.975314),
    ("C", 50.0, 50.0, 0.20, 0.25, 0.025, 0.050): (1.824346,  0.488864),
    ("P", 50.0, 50.0, 0.20, 0.25, 0.025, 0.050): (2.133931, -0.498714),
    ("C", 60.0, 50.0, 0.20, 0.25, 0.025, 0.050): (9.651571, 0.952921),
    ("P", 60.0, 50.0, 0.20, 0.25, 0.025, 0.050): (0.085377, -0.034657),
    # --- long dated, positive carry
    ("C", 40.0, 50.0, 0.20, 1.00, 0.005, 0.010): (0.559848, 0.147521),
    ("P", 40.0, 50.0, 0.20, 1.00, 0.005, 0.010): (10.708478, -0.842529),
    ("C", 50.0, 50.0, 0.20, 1.00, 0.005, 0.010): (3.830202, 0.524620),
    ("P", 50.0, 50.0, 0.20, 1.00, 0.005, 0.010): (4.078334, -0.465430),
    ("C", 60.0, 50.0, 0.20, 1.00, 0.005, 0.010): (10.767967, 0.829743),
    ("P", 60.0, 50.0, 0.20, 1.00, 0.005, 0.010): (1.115601, -0.160306),
}


###############################################################################
class UnitTest(unittest.TestCase):
    # -------------------------------------------------------------------------
    def test_opt_BS(self):
        for parms, (premium, delta) in BLACK_SCHOLES_REF.items():
            p = opt_BS(*parms, ret_type="Premium")
            d = opt_BS(*parms, ret_type="Delta")
            self.assertAlmostEqual(p, premium, 6)
            self.assertAlmostEqual(d, delta, 6)

    # -------------------------------------------------------------------------
    def test_american_pricers(self):
        spot, vol, rd = 100.0, 0.25, 0.05

        # --- test two scenarios for rf (i.e. smaller and larger than rd)
        for rf in (0.03, 0.07):
            for n_days in (60, 120, 255, 510):
                term = n_days / 255.0
                for strike in (75.0, 100, 125):
                    # --- binomial tree: used as benchmark
                    bt_c = opt_BT("C", spot, strike,
                                  vol, term, rd, rf, "A", n_steps=1020)
                    bt_p = opt_BT("P", spot, strike,
                                  vol, term, rd, rf, "A", n_steps=1020)

                    baw_c = opt_BAW("C", spot, strike, vol, term, rd, rf)
                    baw_p = opt_BAW("P", spot, strike, vol, term, rd, rf)

                    c_err = abs(bt_c / baw_c - 1.0)
                    p_err = abs(bt_p / baw_p - 1.0)

                    # --- BAW: test for 4% convergence level (2% can be
                    #          achieved increasing n_steps to 4080)
                    self.assertLessEqual(c_err, 0.04)
                    self.assertLessEqual(p_err, 0.04)

                    jz_c = opt_JZ("C", spot, strike, vol, term, rd, rf)
                    jz_p = opt_JZ("P", spot, strike, vol, term, rd, rf)

                    c_err = abs(bt_c / jz_c - 1.0)
                    p_err = abs(bt_p / jz_p - 1.0)

                    # --- J-Z: test for 1% convergence level
                    self.assertLessEqual(c_err, 0.01)
                    self.assertLessEqual(p_err, 0.01)

    # -------------------------------------------------------------------------
    def test_implied_vols(self):
        strike = 100.0
        rd = 0.05
        for vol in (0.12, 0.30, 0.50):
            for term in (0.25, 0.50, 0.75):
                for rf in (0.00, 0.05, 0.10):
                    for spot in (95.0, 100.0, 105.0):
                        eu_P = opt_BS("P",  spot, strike, vol, term, rd, rf)
                        am_P = opt_BAW("P",  spot, strike, vol, term, rd, rf)
                        eu_C = opt_BS("C", spot, strike, vol, term, rd, rf)
                        am_C = opt_BAW("C", spot, strike, vol, term, rd, rf)

                        iv_eu_P = implied_vol("P", eu_P, spot, strike,
                                              term, rd, rf, american=False)
                        iv_am_P = implied_vol("P", am_P, spot, strike,
                                              term, rd, rf, american=True)
                        iv_eu_C = implied_vol("C", eu_C, spot, strike,
                                              term, rd, rf, american=False)
                        iv_am_C = implied_vol("C", am_C, spot, strike,
                                              term, rd, rf, american=True)

                        self.assertAlmostEqual(iv_eu_P, vol, 3)
                        self.assertAlmostEqual(iv_am_P, vol, 3)
                        self.assertAlmostEqual(iv_eu_C, vol, 3)
                        self.assertAlmostEqual(iv_am_C, vol, 3)


if __name__ == "__main__":
    from onyx.core.utils.unittest import UseEphemeralDbs
    with UseEphemeralDbs():
        unittest.main(failfast=True)



###############################################################################
#
#   Copyright: (c) 2015 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

from onyx.core import Structure, ValueType, UfoBase
from onyx.core import ReferenceField,  StructureField, StringField

from ..risk.decorators import WithRiskValueTypes

__all__ = ["Portfolio"]


###############################################################################
@WithRiskValueTypes
class Portfolio(UfoBase):
    """
    Class used to represent a Portfolio (i.e. a collection of books or other
    sub-portfolios).
    """
    Denominated = ReferenceField(obj_type="Currency")
    DisplayName = StringField()
    Children = StructureField()

    # -------------------------------------------------------------------------
    @ValueType()
    def Books(self, graph):
        """
        Return the list of of all books that are children of this portfolio.
        """
        books = set()
        for child in graph(self, "Children"):
            books.update(graph(child, "Books"))
        return books

    # -------------------------------------------------------------------------
    @ValueType()
    def Leaves(self, graph):
        leaves = Structure()
        for child, qty in graph(self, "Children").items():
            leaves += qty*graph(child, "Leaves")
        return leaves

    # -------------------------------------------------------------------------
    @ValueType()
    def MktValUSD(self, graph):
        mtm = 0.0
        for leaf, qty in graph(self, "Leaves").items():
            mtm += qty*graph(leaf, "MktValUSD")
        return mtm

    # -------------------------------------------------------------------------
    @ValueType()
    def MktVal(self, graph):
        denominated = graph(self, "Denominated")
        spot_fx = graph("{0:3s}/USD".format(denominated), "Spot")
        return graph(self, "MktValUSD") / spot_fx


# -----------------------------------------------------------------------------
def prepare_for_test():
    from onyx.core import AddIfMissing
    from . import ufo_currency
    from . import ufo_currency_cross

    ufo_currency.prepare_for_test()
    ufo_currency_cross.prepare_for_test()

    ports = [
        AddIfMissing(Portfolio(Name="TEST_PORTFOLIO",
                               Denominated="USD",
                               Children=Structure({
                                   "BOOK1": 1.0,
                                   "BOOK2": 1.0}))),
    ]

    return [port.Name for port in ports]



###############################################################################
#
#   Copyright: (c) 2015 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

from onyx.core import RDate, Structure, DateOffset
from onyx.core import GraphNodeVt, GetVal
from onyx.core import IntField, FloatField, DateField
from onyx.core import StringField, SelectField, ReferenceField

from .system.tradable_api import AgingTradableObj, AddByInference
from .system.tradable_api import HashStoredAttrs

from ..system.ufo_forward_cash import ForwardCash
from .ufo_commod_forward import CommodForward
from .ufo_commod_nrby import AVERAGING_TYPES


###############################################################################
class CommodSwap(AgingTradableObj):
    """
    Tradable class that represents a commodity swap contract.
    """
    Asset = ReferenceField(obj_type="CommodAsset")
    AvgStartDate = DateField()
    AvgEndDate = DateField()
    AvgType = SelectField(options=AVERAGING_TYPES)
    RollType = IntField(default=0)
    Quantity = IntField()
    FixedPrice = FloatField()
    Denominated = ReferenceField(obj_type="Currency", default="USD")
    PaymentRule = StringField(default="+5b")

    # -------------------------------------------------------------------------
    @GraphNodeVt("Property")
    def Leaves(self, graph):
        asset = graph(self, "Asset")
        qty = graph(self, "Quantity")
        price = graph(self, "FixedPrice")
        sd = graph(self, "AvgStartDate")
        ed = graph(self, "AvgEndDate")
        one_month = RDate("+1m")
        den = graph(self, "Denominated")
        payment_rule = graph(self, "PaymentDateRule")
        cal = graph(asset, "HolidayCalendar")
        avg_type = graph(self, "AvgType")
        roll_type = graph(self, "RollType")

        # --- settlement on a monthly basis
        leaves = Structure()
        while sd < ed:
            mth_ed = min(ed, DateOffset(sd, "+e"))
            fwd_info = {
                "Asset": asset,
                "AvgStartDate": sd,
                "AvgEndDate": mth_ed,
                "AvgType": avg_type,
                "RollType": roll_type,
            }
            sec = AddByInference(CommodForward(**fwd_info), True)
            leaves[sec.Name] = qty

            # --- cash leg
            cash_info = {
                "Currency": den,
                "PaymentDate": DateOffset(mth_ed, payment_rule, cal),
            }
            sec = AddByInference(ForwardCash(**cash_info), True)
            leaves[sec.Name] = -qty*price

            sd += one_month

        return leaves

    # -------------------------------------------------------------------------
    @GraphNodeVt("Property")
    def UndiscountedValue(self, graph):
        fx = graph("{0:3s}/USD".format(graph(self, "Denominated")), "Spot")
        return graph(self, "UndiscountedValueUSD") / fx

    # -------------------------------------------------------------------------
    @GraphNodeVt("Property")
    def MktVal(self, graph):
        fx = graph("{0:3s}/USD".format(graph(self, "Denominated")), "Spot")
        return graph(self, "MktValUSD") / fx

    # -------------------------------------------------------------------------
    @GraphNodeVt("Property")
    def UndiscountedValueUSD(self, graph):
        val = 0.0
        for sec, qty in graph(self, "Leaves").items():
            val += qty*graph(sec, "UndiscountedValueUSD")
        return val

    # -------------------------------------------------------------------------
    @GraphNodeVt("Property")
    def MktValUSD(self, graph):
        val = 0.0
        for sec, qty in graph(self, "Leaves").items():
            val += qty*graph(sec, "MktValUSD")
        return val

    # -------------------------------------------------------------------------
    @GraphNodeVt("Property")
    def ExpirationDate(self, graph):
        securities = graph(self, "Leaves").keys()
        return max([GetVal(sec, "ExpirationDate") for sec in securities])

    # -------------------------------------------------------------------------
    @GraphNodeVt("Property")
    def NextTransactionDate(self, graph):
        securities = graph(self, "Leaves").keys()
        return max([GetVal(sec, "NextTransactionDate") for sec in securities])

    # -------------------------------------------------------------------------
    @GraphNodeVt("Property")
    def NextTransactionEvent(self, graph):
        return "Settlement"

    # -------------------------------------------------------------------------
    @GraphNodeVt("Property")
    def NextTransactionSecurity(self, graph):
        return None

    # -------------------------------------------------------------------------
    @property
    def ImpliedName(self, graph):
        mkt = GetVal(self.Asset, "Market")
        sym = GetVal(self.Asset, "Symbol")
        return ("CmdSWP {0:s} {1:s} {2:8s} "
                "{0:2d").format(mkt, sym, HashStoredAttrs(self, 8))



###############################################################################
#
#   Copyright: (c) 2015 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

from onyx.core import Date, Curve, Knot, HlocvCurve
from onyx.core import CalcTerm, DateOffset, Interpolate
from onyx.core import ObjDbTransaction, DelObj, ObjNotFound
from onyx.core import TsDbQuery, TsDbGetCurve, TsNotFound
from onyx.core import GetVal, ValueType, RetainedFactory
from onyx.core import StringField, DictField, ReferenceField
from onyx.core import CurveField, BoolField

from ..system.ufo_asset import Asset

import numpy as np
import math

__all__ = ["EquityAsset"]

# --- map drop-down values to value types (used by edit screens)
CRV_CHOICES = {
    "Adjusted": "GetCurve",
    "Marks": "GetMarks",
    "Dividends": "GetDividends",
}


###############################################################################
class EquityAsset(Asset):
    """
    class used to represent an equity asset and to provide access to its
    curve of historical prices and price fixes.
    """
    Tickers = DictField()
    Isin = StringField()
    Sedol = StringField()
    Country = ReferenceField(obj_type="Category")
    Region = ReferenceField(obj_type="Category")
    Sector = ReferenceField(obj_type="Category")
    Subsector = ReferenceField(obj_type="Category")
    TimeSeries = StringField()
    Marks = StringField()
    Dividends = StringField()
    SharesOutCrv = CurveField(default=Curve())
    IsProxySecurity = BoolField(default=False)
    ReferenceIndex = ReferenceField(obj_type="EquityIndex")

    name_fmt = "EQ {0:s} {1:s}"

    # -------------------------------------------------------------------------
    def __post_init__(self):
        super().__post_init__()

        exchange_code = GetVal(self.Exchange, "Code")

        self.Name = self.name_fmt.format(self.Symbol, exchange_code)
        self.Country = self.Country or GetVal(self.Exchange, "Country")
        self.Region = self.Region or GetVal(self.Exchange, "Region")

        # --- name of the time-series that store historical prices,
        #     historical closes (marks) and dividends
        fields = self.Symbol, exchange_code
        self.TimeSeries = "EQ-TS {0:s} {1:s}".format(*fields)
        self.Marks = "EQ-MKS {0:s} {1:s}".format(*fields)
        self.Dividends = "EQ-DIV {0:s} {1:s}".format(*fields)

    # -------------------------------------------------------------------------
    @ValueType()
    def Ticker(self, graph):
        return graph(self, "Tickers")[graph("Settings", "Platform")]

    # -------------------------------------------------------------------------
    @ValueType()
    def UniqueId(self, graph):
        sym = graph(self, "Symbol")
        code = graph(graph(self, "Exchange"), "Code")
        return "{0:s} {1:2s}".format(sym, code)

    # -------------------------------------------------------------------------
    @RetainedFactory()
    def Spot(self, graph):
        """
        Return the official close value as of MktDataDate (or the most recent
        close if ForceStrict is False) in the Denominated currency.
        """
        return graph(graph(self, "Marks"), "Price")*graph(self, "Multiplier")

    # -------------------------------------------------------------------------
    @ValueType()
    def Last(self, graph):
        """
        Return the knot with the most recent close value (irrespective of
        MktDataDate) in the Denominated currency.
        """
        marks = graph(self, "Marks")
        date, value = graph(marks, "LastKnot", date=Date.high_date())
        return Knot(date, value*graph(self, "Multiplier"))

    # -------------------------------------------------------------------------
    @ValueType("PropSubGraph")
    def GetCurve(self, graph, start=None, end=None, field=None, adj=True):
        name = graph(self, "TimeSeries")
        try:
            crv = TsDbGetCurve(name, start, end, "HLOCV", field)
        except TsNotFound:
            return HlocvCurve() if field is None else Curve()

        if adj:
            return graph(graph(self, "Dividends"), "AdjustCurve")(crv)
        else:
            return crv

    # -------------------------------------------------------------------------
    @ValueType("PropSubGraph")
    def GetMarks(self, graph, start=None, end=None):
        return graph(graph(self, "Marks"), "PrcFixCurve", start, end)

    # -------------------------------------------------------------------------
    @ValueType("PropSubGraph")
    def GetDividends(self, graph, start=None, end=None):
        return graph(graph(self, "Dividends"), "DvdCurve", start, end)

    # -------------------------------------------------------------------------
    @ValueType()
    def DividendForecast(self, graph):
        """
        Include dividends paid within the last 12 months and pick the most
        recent as forcasted value.
        """
        mdd = graph("Database", "MktDataDate")
        div = graph(self, "GetDividends").crop(DateOffset(mdd, "-12m"))
        val = Interpolate(div, mdd, method="Step")

        rules = ["+1y-0b", "+2y-0b", "+3y-0b"]
        cal = graph(self, "HolidayCalendar")
        dts = [[DateOffset(d, r, cal) for r in rules] for d in div.dates]
        dts = {item for sublist in dts for item in sublist}
        vls = val*np.ones(len(dts))

        return Curve(dts, vls)

    # -------------------------------------------------------------------------
    @ValueType("Callable")
    def DividendYield(self, graph, start, end):
        """
        DividendYield, calculated by including all dividends that are expected
        to be paid between start and end date and annualizing with continuous
        compounding.
        """
        if start <= end:
            return 0.0

        fcast = graph(self, "DividendForecast").crop_values(start, end)
        tot_dvd = fcast.sum()*graph(self, "Multiplier")
        spot = graph(self, "Spot")

        return math.log(tot_dvd/spot + 1.0) / CalcTerm(start, end)

    # -------------------------------------------------------------------------
    @ValueType("PropSubGraph")
    def NextExDateActions(self, graph, start=None):
        if start is None:
            start = graph("Database", "PositionsDate")
        return graph(graph(self, "Dividends"), "NextExDateActions", start)

    # -------------------------------------------------------------------------
    @ValueType()
    def SharesOutstanding(self, graph):
        return Interpolate(graph(self, "SharesOutCrv"),
                           graph("Database", "MktDataDate"), method="Step")

    # -------------------------------------------------------------------------
    #  Auxiliary VTs used by the edit screen

    @RetainedFactory(default="Adjusted")
    def Selection(self, graph):
        pass

    @ValueType()
    def DisplayCurve(self, graph):
        return graph(self, CRV_CHOICES[graph(self, "Selection")])

    # -------------------------------------------------------------------------
    def delete(self):
        # --- delete timeseries
        TsDbQuery("DELETE FROM HlocvCurves WHERE "
                  "Name = %s", parms=(self.TimeSeries,))

        # --- delete price fixes
        with ObjDbTransaction("cleanup", "SERIALIZABLE"):
            for obj_name in (self.Marks, self.Dividends):
                try:
                    DelObj(obj_name)
                except ObjNotFound:
                    # --- this is fine, ignore exception
                    continue


# -----------------------------------------------------------------------------
def prepare_for_test():
    from onyx.core import AddIfMissing, EvalBlock
    from ..system.ufo_price_fix import PriceFix

    from ..system import ufo_database
    from ..system import ufo_settings
    from ..system import ufo_holiday_calendar
    from ..system import ufo_currency
    from ..system import ufo_exchange
    from ..system import ufo_price_fix

    ufo_database.prepare_for_test()
    ufo_settings.prepare_for_test()
    ufo_holiday_calendar.prepare_for_test()
    ufo_currency.prepare_for_test()
    ufo_exchange.prepare_for_test()
    ufo_price_fix.prepare_for_test()

    ibm_info = {
        "Symbol": "IBM",
        "Exchange": "XNYS Exchange",
        "Tickers": {"Bloomberg": "IBM"},
        "Multiplier": 1.00,
    }

    sec = AddIfMissing(EquityAsset(**ibm_info))
    prc_fix = AddIfMissing(PriceFix(Name=GetVal(sec, "Marks")))

    with EvalBlock() as eb:
        eb.change_value("Database", "ArchivedOverwritable", True)
        prc_fix.set_dated("Price", Date.today(), 100.0)

    ng_info = {
        "Symbol": "NG/",
        "Exchange": "XLON Exchange",
        "Tickers": {
            "Bloomberg": "NG/",
            "Google": "NG",
            "Yahoo": "NG",
            "WSJ": "NG."
        },
        "Multiplier": 0.01,
    }

    sec = AddIfMissing(EquityAsset(**ng_info))
    prc_fix = AddIfMissing(PriceFix(Name=GetVal(sec, "Marks")))
    with EvalBlock() as eb:
        eb.change_value("Database", "ArchivedOverwritable", True)
        prc_fix.set_dated("Price", Date.today(), 900.0)



###############################################################################
#
#   Copyright: (c) 2015 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

from .black_scholes import opt_BS

from numpy.polynomial import polynomial, laguerre, hermite_e

import numpy as np
import math

__all__ = ["MC_vanilla", "MC_american"]


# -----------------------------------------------------------------------------
def MC_vanilla(prices, payoff, rd, rf, term, bstrap=False):
    """
    Description:
        Monte-Carlo pricer for payoffs that only depend on terminal (marginal)
        distribution of prices (european style exercise).
    Inputs:
        prices - ensemble of terminal prices
        payoff - scalar payoff function
        term   - time to expiration (units must be consistent with rd and rf)
        rd     - domestic risk-free interest rate
        rf     - foreign risk-free interest rate (or dividend yield)
        bstrap - if Trues, use bootstrap resampling to get better estimate of
                 monte-carlo error.
                 NB: current implementation seems to improve estimates very
                 little...
    Returns:
         A tuple of floats with premium and its error.
    """
    # --- vectorize payoff function
    payoff = np.vectorize(payoff, otypes=[float])

    # --- discount factor
    df = math.exp(-rd*term)

    # --- apply payoff to price paths
    samples = payoff(prices)

    if bstrap:
        n = samples.shape[0]
        bstrap_samples = []
        for _ in range(min(n, 1000)):
            idxs = np.random.random_integers(0, n - 1, n)
            bstrap_samples.append(samples[idxs].mean())
        return df*np.mean(bstrap_samples), df*np.std(bstrap_samples)
    else:
        return (df*samples.mean(),
                df*samples.std() / math.sqrt(samples.shape[0] - 1.0))


# -----------------------------------------------------------------------------
def MC_american(paths, payoff, term, rd, rf,
                basis="POLY", full_distr=False, **kwds):
    """
    Description:
        Monte-Carlo pricer for payoffs with american-style exercise.
    Inputs:
        paths      - ensemble of prices paths
        payoff     - scalar payoff function
        term       - time to expiration (units must be consistent with rd, rf)
        rd         - domestic risk-free interest rate
        rf         - foreign risk-free interest rate (or dividend yield)
        basis      - basis set used to fit extrinsic continuation value: choose
                     among "POLY", "LAGUERRE", "HERMITE"
        full_distr - if True, returns the whole distribution, rather than the
                     mean
        kwds       - extra named arguments used for pricing an equivalent
                     european option for variance reduction.
                     Required parameters are:
                        - opt_type - option type (CALL, PUT)
                        - spot     - current spot price
                        - strike   - strike of the option
                        - vol      - price volatility (standard deviation of
                                     log returns)
    Returns:
        A tuple of floats (premium and its error) or a numpy.array if
        full_distr is True.
    """
    # --- vectorize payoff function
    payoff = np.vectorize(payoff, otypes=[float])

    # --- generate function used to fit the extrinsic continuation value
    if basis.upper() == "POLY":
        def extrcontval(x, y, deg):
            # --- Vandermonde matrix of powers of x
            B = polynomial.polyvander(x, deg).T
            return np.dot(B.T,
                          np.dot(np.linalg.pinv(np.dot(B, B.T)),
                                 np.dot(B, y)))

    elif basis.upper() == "LAGUERRE":
        def extrcontval(x, y, deg):
            # --- Vandermonde matrix of weighted Laguerre polynomials of
            #     x plus constant
            wx = np.exp(-0.5*x)
            B = np.vstack((np.ones_like(x), wx*laguerre.lagvander(x, deg).T))
            return np.dot(B.T,
                          np.dot(np.linalg.pinv(np.dot(B, B.T)),
                                 np.dot(B, y)))

    elif basis.upper() == "HERMITE":
        def extrcontval(x, y, deg):
            # --- Vandermonde matrix of weighted Hermite polynomials of
            #     x plus constant
            wx = np.exp(-0.5*x*x)
            B = np.vstack((np.ones_like(x),
                           wx*hermite_e.hermevander(x, deg).T))
            return np.dot(B.T,
                          np.dot(np.linalg.pinv(np.dot(B, B.T)),
                                 np.dot(B, y)))

    else:
        NameError("Unrecognized basis {0:s}".format(basis))

    # --- index of last time step
    last = paths.shape[0] - 1

    # --- one-step discount factor
    df = math.exp(-rd * term / float(last + 1))

    # --- walk backward: at each step cv contains the PV at that step of
    #     the cashflows conditional to not having exercised the option yet.
    for k in range(last, 0, -1):
        if k == last:
            # --- at expiration the continuation value is zero: always
            #     exercise the option if ITM
            cf = payoff(paths[k,:])
        else:
            # --- cash-flow at current time-step:
            #     1) if a path is OTM keep the option: current cash-flow is
            #        the PV of the one step ahead cash-flow.
            #     2) if a path is ITM:
            #        2.1) if expected continuation value is more than
            #             intrinsic keep the option: current cash-flow is
            #             the PV of the one step ahead cash-flow.
            #        2.2) otherwise exercise: current cash-flow is the
            #             intrinsic value.

            # --- start by calculating PV of the one step ahead cash-flow
            cf *= df

            # --- intrinsic value at current step
            intrinsic = payoff(paths[k,:])

            # --- indexes of in the money paths
            itm = np.nonzero(np.ravel(intrinsic > 0.0))[0]

            # --- for the ITM paths (if any), compare expected conditional
            #     continuation value to the intrinsic value
            if len(itm):
                # --- fit extrinsic continuation value
                x = paths[k,itm]
                y = cf[itm] - intrinsic[itm]

                # --- scale x to improve fit's condition number
                xmin, xmax = x.min(), x.max()
                if xmax > xmin:
                    x = (x - xmin) / (xmax - xmin)

                degree  = max(1, int(math.log(len(itm), 4)))
                cf[itm] = np.where(extrcontval(x, y, degree) > 0.0,
                                   cf[itm], intrinsic[itm])

    # --- try to get extra parameters needed to use the exact price of an
    #     equivalent european option as control variate
    try:
        opt_type = kwds["opt_type"]
        spot = kwds["spot"]
        strike = kwds["strike"]
        vol = kwds["vol"]

        # --- PV of payoffs for the european option
        euro_mc = math.exp(-rd*term)*payoff(paths[-1,:])
        euro_bs = opt_BS(opt_type, spot, strike, vol, term, rd, rf)

        # --- determine the coefficient that minimizes the variance of the
        #     estimate
        covmat = np.cov(cf, euro_mc)
        coeff = covmat[0,1] / covmat[1,1]    

        cf += coeff*(euro_bs - euro_mc)

    except KeyError:
        pass

    if full_distr:
        return cf
    else:
        return cf.mean(), cf.std() / math.sqrt(cf.shape[0] - 1.0)



###############################################################################
#
#   Copyright: (c) 2015 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

from onyx.core import GetVal, Structure, OnyxTestCase
from onyx.risk import ufo_risk

import unittest


###############################################################################
class UnitTest(OnyxTestCase):
    # -------------------------------------------------------------------------
    def setUp(self):
        super().setUp()
        self.book, *_ = ufo_risk.prepare_for_test()

    # -------------------------------------------------------------------------
    def compareStructures(self, first, second, places):
        self.assertEqual(first.keys(), second.keys())
        for key, value in first.items():
            self.assertAlmostEqual(value, second[key], places)

    # -------------------------------------------------------------------------
    def test_MktVal(self):
        self.assertAlmostEqual(GetVal(self.book, "MktVal"), 200.0, 8)

    # -------------------------------------------------------------------------
    def test_MktValUSD(self):
        self.assertAlmostEqual(GetVal(self.book, "MktValUSD"), 300.0, 8)

    # -------------------------------------------------------------------------
    def test_Deltas(self):
        ref_deltas = Structure([("EQ NG/ LN", -2000.0)])
        self.compareStructures(GetVal(self.book, "Deltas"), ref_deltas, 4)

    # -------------------------------------------------------------------------
    def test_Exposures(self):
        self.assertAlmostEqual(GetVal(self.book, "GrossExposure"), 18000.0, 4)
        self.assertAlmostEqual(GetVal(self.book, "NetExposure"), -18000.0, 4)

    # -------------------------------------------------------------------------
    def test_FX(self):
        GetVal(self.book, "FxExposures")
        ref_fx = Structure([
            ("EUR/USD", 11869.565217391306),
            ("GBP/USD", -8900.0)
        ])
        self.compareStructures(GetVal(self.book, "FxExposures"), ref_fx, 8)


if __name__ == "__main__":
    from onyx.core.utils.unittest import UseEphemeralDbs
    with UseEphemeralDbs():
        unittest.main(failfast=True)



###############################################################################
#
#   Copyright: (c) 2015 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

from onyx.core import StringField, DictField

from .ufo_book import Book

__all__ = ["Broker"]


###############################################################################
class Broker(Book):
    """
    Class used to represent a broker.
    """
    BookType = StringField(default="Broker")
    FeeSchedule = DictField()


# -----------------------------------------------------------------------------
def prepare_for_test():
    from onyx.core import AddIfMissing

    brokers = [
        AddIfMissing(Broker(Name="TEST_BROKER")),
        AddIfMissing(Broker(Name="BROKER-SUSPENSE")),
    ]

    return [broker.Name for broker in brokers]



###############################################################################
#
#   Copyright: (c) 2015 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

from onyx.core import Date, Date2LYY, DateOffset
from onyx.core import Curve, HlocvCurve
from onyx.core import AddObj, GetObj, DelObj, ObjNotFound, ObjExists
from onyx.core import ObjDbTransaction
from onyx.core import TsDbGetCurve, TsNotFound
from onyx.core import GetVal, ValueType
from onyx.core import StringField, SelectField, IntField, FloatField
from onyx.core import SetField, DictField

from ..risk.timeseries import prices_for_risk
from ..system.ufo_asset import WithAssetAttributes
from .ufo_commod_contract import CommodCnt

from dateutil import relativedelta

import logging

__all__ = ["CommodAsset"]

VALID_UNITS = ["BBL", "MT", "MWh", "Therm", "MMBTU", "Day"]

logger = logging.getLogger(__name__)


###############################################################################
class CommodAsset(WithAssetAttributes):
    """
    This class used to represent commodity assets and to get access to the
    underlying contracts.
    Despite its name, this is not an Asset as it doesn't expose the full Asset
    interface: this because the underlying of a commod tradable is a specific
    commod contract (which is the actual Asset).
    """
    Market = StringField()
    Tickers = DictField()
    Units = SelectField(options=VALID_UNITS)
    TimeSeries = StringField()

    # --- SettDateRule is used to determine the futures settlement date of a
    #     futures contract from its LYY code as follows:
    #         FutSettDate = DateOffset(LYY2Date(DeliveryMonth), SettDateRule)
    SettDateRule = StringField(default="+e")

    # --- OptExpDateRule is used to determine the options expiration date
    #     starting from the settlement date of the futures contract.
    OptExpDateRule = StringField(default="+e")

    CntType = SelectField(options=["CLOSE", "HLOCV"])
    NrbyOffset = IntField()

    ContractSize = FloatField(default=1.0)
    Contracts = SetField(default=set())

    # --- used for attributions
    Country = StringField(default="Commodity")
    Region = StringField(default="Commodity")
    Sector = StringField(default="Commodity")
    Subsector = StringField(default="Commodity")

    # -------------------------------------------------------------------------
    def __post_init__(self):
        super().__post_init__()

        self.Name = "COMMOD {0:s} {1:s}".format(self.Market, self.Symbol)

        # --- calculate the neraby offset based on the settlement date rule and
        #     an arbitrary reference date
        ref = DateOffset(Date.today(), "+0J")
        di = DateOffset(ref, self.SettDateRule, self.HolidayCalendar)
        df = DateOffset(ref, "+e", self.HolidayCalendar)
        rd = relativedelta.relativedelta(df, di)

        self.NrbyOffset = rd.months + 12*rd.years
        if self.NrbyOffset < 0:
            raise RuntimeError("Negative NrbyOffset?!? Check SettDateRule.")

        # --- name of the time-series where historical cash prices are stored
        self.TimeSeries = "CMD-TS {0:s} {1:s}".format(self.Market, self.Symbol)

    # -------------------------------------------------------------------------
    @ValueType()
    def Ticker(self, graph):
        return graph(self, "Tickers")[graph("Settings", "Platform")]

    # -------------------------------------------------------------------------
    @ValueType("Callable")
    def GetContract(self, graph, del_mth):
        """
        Description:
            Return the CommodCnt object for a specific delivery month.
        Inputs:
            del_mth - the delivery month in LYY format (as in Z11)
        Returns:
            A string.
        """
        if del_mth in graph(self, "Contracts"):
            market = graph(self, "Market")
            symbol = graph(self, "Symbol")
            return CommodCnt.get_name(market, symbol, del_mth)
        else:
            raise NameError("Contract {0:s} not found in the set of "
                            "Contracts of {1:s}".format(del_mth, self.Name))

    # -------------------------------------------------------------------------
    @ValueType("PropSubGraph")
    def GetCurve(self, graph, start=None, end=None, field=None):
        name = graph(self, "TimeSeries")
        if graph(self, "CntType") == "HLOCV":
            try:
                return TsDbGetCurve(name, start, end, "HLOCV", field)
            except TsNotFound:
                return HlocvCurve() if field is None else Curve()
        else:
            try:
                return TsDbGetCurve(name, start, end)
            except TsNotFound:
                return Curve()

    # -------------------------------------------------------------------------
    @ValueType("Callable")
    def PricesForRisk(self, graph, start, end):
        return prices_for_risk(graph(self, "Name"), start, end, strict=False)

    # -------------------------------------------------------------------------
    @ValueType("Callable")
    def ActiveByDate(self, graph, date):
        """
        Return the active contract (first nearby) as of a given date.
        """
        cal = graph(self, "HolidayCalendar")
        sdr = graph(self, "SettDateRule")

        rule = "+{0:d}m+0J".format(graph(self, "NrbyOffset"))
        cnt_dt = DateOffset(date, rule)

        if date > DateOffset(cnt_dt, sdr, cal):
            cnt_dt = DateOffset(cnt_dt, "+1m")

        return Date2LYY(cnt_dt)

    # -------------------------------------------------------------------------
    def add_contract(self, del_mth, tickers):
        """
        Description:
            Add a new contract or return the existing one.
        Inputs:
            cnt_mth - the contract month in LYY format
            tickers - a dictionary of tickers
        Returns:
            The contract's name.
        """
        if del_mth in self.Contracts:
            return GetVal(self, "GetContract", del_mth)
        else:
            info = {
                "CommodAsset": self.Name,
                "DeliveryMonth": del_mth,
                "Tickers": tickers,
                "RiskProxy": self.Name,
            }
            cnt_obj = CommodCnt(**info)
            try:
                AddObj(cnt_obj)
            except ObjExists:
                obj = GetObj(cnt_obj.Name)
                if obj != cnt_obj:
                    raise RuntimeError("StoredAttrs of existing contract "
                                       "{0:s} don't match those of parent "
                                       "CommodAsset".format(obj.Name))

            self.Contracts.add(del_mth)

        return cnt_obj.Name

    # -------------------------------------------------------------------------
    def delete(self):
        mkt = self.Market
        sym = self.Symbol
        with ObjDbTransaction("deleting contracts", "SERIALIZABLE"):
            # --- conversion to tuple is needed because the delete method of a
            #     contract removes such contract from the set of contracts of
            #     the asset.
            for cnt in tuple(self.Contracts):
                cnt_name = CommodCnt.get_name(mkt, sym, cnt)
                try:
                    DelObj(cnt_name)
                except ObjNotFound:
                    logger.waring("contract "
                                  "{0:s} not found ?!?".format(cnt_name))


# -----------------------------------------------------------------------------
def prepare_for_test():
    from onyx.core import AddIfMissing

    from ..system import ufo_database
    from ..system import ufo_settings
    from ..system import ufo_holiday_calendar
    from ..system import ufo_currency
    from ..system import ufo_exchange

    ufo_database.prepare_for_test()
    ufo_settings.prepare_for_test()
    ufo_holiday_calendar.prepare_for_test()
    ufo_currency.prepare_for_test()
    ufo_exchange.prepare_for_test()

    eua_info = {
        "Symbol": "EUA",
        "Market": "CO2",
        "Exchange": "IFEN Exchange",
        "Units": "MT",
        "CntType": "CLOSE",
        "SettDateRule": "+e-M+4d-4b-M+4d-4b-M",
        "OptExpDateRule": "-3b",
        "Tickers": {"Bloomberg": "MO1"},
        "Multiplier": 1.0,
        "ContractSize": 1000.0,
    }

    AddIfMissing(CommodAsset(**eua_info))



###############################################################################
#
#   Copyright: (c) 2015 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

from onyx.core import Structure, TsDbGetRowBy, ValueType, ReferenceField
from onyx.core import FloatField, BoolField, DateField, StringField

from .tradable_api import AgingTradableObj, AddByInference, HashStoredAttrs
from .ufo_forward_cash import ForwardCash
from .ufo_cash_balance import CashBalance


###############################################################################
class FxForward(AgingTradableObj):
    """
    Tradable class that represents a FX forward (in either a deliverable or
    non-deliverable currency).
    """
    # --- this is the currency received forward
    CurrencyRec = ReferenceField(obj_type="Currency")

    # --- this is the currency delivered forward
    CurrencyDel = ReferenceField(obj_type="Currency")

    # --- the agreed forward fx rate: this determines the amount of CurrencyDel
    FxRate = FloatField()

    # --- this represents the delivery date for deliverable forwards and the
    #     fixing date for non-deliverable forwards
    DeliveryDate = DateField()

    # --- False for a non-deliverable forward (NDF). Only affects settlement.
    Deliverable = BoolField(default=True)

    # --- this field is only used for non-deliverable forwards. The relevant
    #     calendar is that of the delivered currency.
    SettDateRule = StringField(default="+2b")

    # -------------------------------------------------------------------------
    @ValueType()
    def Leaves(self, graph):
        ccy_rec = graph(self, "CurrencyRec")
        ccy_del = graph(self, "CurrencyDel")
        mkd_dt = graph("Database", "MktDataDate")
        del_dt = graph(self, "DeliveryDate")

        fwd_ccy1 = ForwardCash(Currency=ccy_rec, PaymentDate=del_dt)
        fwd_ccy1 = AddByInference(fwd_ccy1, in_memory=True)
        fwd_ccy2 = ForwardCash(Currency=ccy_del, PaymentDate=del_dt)
        fwd_ccy2 = AddByInference(fwd_ccy2, in_memory=True)

        # --- if the  fx-forward is non deliverable and we are pricing after
        #     the fixing date, use spot-fx to calculate the net amount of
        #     CurrencyRec
        if not graph(self, "Deliverable") and mkd_dt >= del_dt:
            fixing_rate = graph(self, "FixingRate")

            return Structure({
                fwd_ccy1.Name: 1.0 - graph(self, "FxRate") / fixing_rate,
            })

        if fwd_ccy1.Name == fwd_ccy2.Name:
            return Structure({fwd_ccy1.Name: 1.0 - graph(self, "FxRate")})
        else:
            return Structure({
                fwd_ccy1.Name: 1.0,
                fwd_ccy2.Name: -graph(self, "FxRate"),
            })

    # -------------------------------------------------------------------------
    @ValueType()
    def FixingRate(self, graph):
        mdd = graph("Database", "MktDataDate")
        fxd = graph(self, "DeliveryDate")

        # --- before the fixing date FixingRate is not defined nor needed
        if mdd < fxd:
            return None

        ccy_rec = graph(self, "CurrencyRec")
        ccy_del = graph(self, "CurrencyDel")
        strict = graph("Database", "ForceStrict")

        rec_cross = graph("{0:3s}/USD".format(ccy_rec), "FxTimeSeries")
        rec_spot = TsDbGetRowBy("Curves", rec_cross, fxd, strict)[2]

        del_cross = graph("{0:3s}/USD".format(ccy_del), "FxTimeSeries")
        del_spot = TsDbGetRowBy("Curves", del_cross, fxd, strict)[2]

        return rec_spot / del_spot

    # -------------------------------------------------------------------------
    @ValueType()
    def MktValUSD(self, graph):
        mtm_usd = 0.0
        for leaf, qty in graph(self, "Leaves").items():
            mtm_usd += qty*graph(leaf, "MktValUSD")
        return mtm_usd

    # -------------------------------------------------------------------------
    @ValueType()
    def MktVal(self, graph):
        """
        By convenction, the reference currency of an FX Forward is the currency
        that is delivered forward.
        """
        cross = "{0:3s}/USD".format(graph(self, "CurrencyDel"))
        return graph(self, "MktValUSD") / graph(cross, "Spot")

    # -------------------------------------------------------------------------
    @ValueType()
    def ExpirationDate(self, graph):
        return graph(self, "DeliverytDate")

    # -------------------------------------------------------------------------
    @ValueType()
    def NextTransactionDate(self, graph):
        return graph(self, "DeliverytDate")

    # -------------------------------------------------------------------------
    @ValueType()
    def TradeTypes(self, graph):
        mapping = super().TradeTypes
        mapping.update({
            "Settle": "SettleSecurities",
        })
        return mapping

    # -------------------------------------------------------------------------
    @ValueType()
    def ExpectedTransaction(self, graph):
        return "Settle"

    # -------------------------------------------------------------------------
    @ValueType()
    def SettleSecurities(self, graph):
        if graph(self, "Deliverable"):
            cash_rec = CashBalance(Currency=graph(self, "CurrencyRec"))
            cash_del = CashBalance(Currency=graph(self, "CurrencyDel"))
            fixing_rate = 1.0
        else:
            cash_rec = CashBalance(Currency=graph(self, "CurrencyRec"))
            cash_del = cash_rec
            fixing_rate = graph(self, "FixingRate")

        return [
            {"Security": AddByInference(cash_del, in_memory=True),
             "Quantity": 1.0},
            {"Security": AddByInference(cash_rec, in_memory=True),
             "Quantity": -graph(self, "FxRate") / fixing_rate},
        ]

    # -------------------------------------------------------------------------
    @property
    def ImpliedName(self):
        ccy_del = self.CurrencyDel
        ccy_rec = self.CurrencyRec
        del_date = self.DeliveryDate.strftime("%d%m%y")
        mush = HashStoredAttrs(self, 4)
        return ("FX-FWD {0:3s}/{1:3s} {2:6s} {3:4s} "
                "{{0:2d}}").format(ccy_del, ccy_rec, mush, del_date)


# -----------------------------------------------------------------------------
def prepare_for_test():
    from onyx.core import Date, RDate
    from .tradable_api import AddByInference
    from . import ufo_forward_cash

    ufo_forward_cash.prepare_for_test()

    del_date = Date.today() + RDate("+1y")

    securities = []
    for ccy in ["EUR", "GBP"]:
        securities.append(AddByInference(FxForward(CurrencyDel=ccy,
                                                   CurrencyRec="USD",
                                                   FxRate=1.00,
                                                   DeliveryDate=del_date)))

    return [sec.Name for sec in securities]



###############################################################################
#
#   Copyright: (c) 2015 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

from .black_scholes import opt_BS

from math import exp as m_exp, sqrt as m_sqrt
from numpy import exp, where

import numpy as np

__all__ = ["opt_BT"]


# -----------------------------------------------------------------------------
def opt_BT(opt_type, spot, strike,
           vol, term, rd, rf, style="EUROPEAN", n_steps=510):
    """
    Description:
        Black-Scholes option pricer based on Binomial Tree discretization of
        time.
        It can price both european and american exercise-style options.
        Black76 formula for futures is retrived by setting rd = rf = r.
    Inputs:
        opt_type - option type (CALL, PUT). Only first letter is used.
        spot     - current spot price
        strike   - strike of the option
        vol      - price volatility (standard deviation of log returns)
        term     -  time to expiration (units must be consistent with what is
                   used for vol, rd, rf)
        rd       - domestic risk-free interest rate
        rf       - foreign risk-free interest rate (or dividend yield)
        style    - exercise style (EUROPEAN, AMERICAN). Only first letter
                   is used.
        n_steps  - number of discretization time-steps on a per-year basis
                   (default is 510 steps per year, about 2 per biz-day)
    Returns:
        The option's fair value
    """
    # --- convert string arguments to upper-case
    opt_type = opt_type[0].upper()
    style = style[0].upper()

    # --- Calls
    if opt_type == "C":
        payoff = np.vectorize(lambda s: max(s - strike, 0.0), otypes=[float])
    # --- Puts
    elif opt_type == "P":
        payoff = np.vectorize(lambda s: max(strike - s, 0.0), otypes=[float])
    else:
        raise NameError("Unrecognized option type, {0:s}".format(opt_type))

    # --- convert n_steps to total number of steps for the binomial tree
    n_steps = int(n_steps*term)

    dt = term / float(n_steps)
    df = m_exp(-rd*dt)

    drift = (rd - rf - 0.5*vol*vol)
    volsqrtdt = vol*m_sqrt(dt)

    # --- european exercise
    if style == "E":
        for k in range(n_steps, -1, -1):
            if k == n_steps:
                opt = payoff(spot*exp(drift*k*dt +
                                      volsqrtdt*np.linspace(-k, k, k+1)))
            else:
                # --- PV of the one-step-ahead expected value
                opt = df*0.5*(opt[1:] + opt[:-1])

        return opt[0]

    # --- american exercise (it uses a european option as control variate
    #     to reduce error)
    elif style == "A":
        for k in range(n_steps, -1, -1):
            if k == n_steps:
                opt = payoff(spot*exp(drift*k*dt +
                                      volsqrtdt*np.linspace(-k, k, k+1)))
                euro = opt
            else:
                # --- intrinsic value
                intrinsic = payoff(spot*exp(drift*k*dt +
                                            volsqrtdt*np.linspace(-k, k, k+1)))

                # --- continuation value: PV of the one-step-ahead
                #     expected value
                cont = df*0.5*(opt[1:] + opt[:-1])

                # --- option's value is the best of the intrinsic and
                #     continuation values
                opt = where(cont > intrinsic, cont, intrinsic)

                # --- european option (used for control variate)
                euro = df*0.5*(euro[1:] + euro[:-1])

        # --- assume error for the european option can be used as a correction
        #     for the american one. However, always enforce that fair value is
        #     at least the intrinsic value.
        exact_bs = opt_BS(opt_type, spot, strike, vol, term, rd, rf)
        return max(opt[0] + exact_bs - euro[0], payoff(spot))

    else:
        raise NameError("Unrecognized exercise style, {0:s}".format(style))
